/*
 *
 *  Bluetooth OBEX tool
 *
 *  Copyright (C) 2002  Marcel Holtmann <marcel@holtmann.org>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Code Modified by Thomson. Revision 1.0.0
 *
 */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>
#include <libgen.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/rfcomm.h>

#include "glib-ectomy.h"
#include <openobex/obex.h>

#include <sys/wait.h>

volatile int finished = 0;
volatile int success = 0 ;

struct btobex_context_t {
	int fd;
  int input_file ;
  char name[128] ;
  //char full_name[128] ;
  char fifo_path[128] ;
	int fifo_pump_pid;
  obex_t *obexhandle ;
	bdaddr_t bdaddr;
	uint8_t channel;
} btobex_context;

static char inbox_script[128];

gint btobex_srv_receive(struct btobex_context_t *srv, obex_object_t *object, 
    gboolean finished) ;
gint btobex_srv_setpath(struct btobex_context_t *srv, obex_object_t *object) ;


gint btobex_connect(obex_t *handle, gpointer userdata)
{
	struct btobex_context_t *context = userdata;
	struct sockaddr_rc laddr, raddr;
	int s;

	if ((s = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
		printf("Can't create socket. %s (%d)\n", strerror(errno), errno);
		return -1;
	}

	laddr.rc_family = AF_BLUETOOTH;
	bacpy(&laddr.rc_bdaddr, BDADDR_ANY);
	laddr.rc_channel = 0;
	if (bind(s, (struct sockaddr *)&laddr, sizeof(laddr)) < 0) {
		printf("Can't bind socket. %s (%d)\n", strerror(errno), errno);
		close(s);
		return -1;
	}

	raddr.rc_family = AF_BLUETOOTH;
	bacpy(&raddr.rc_bdaddr, &context->bdaddr);
	raddr.rc_channel = context->channel;
	if (connect(s, (struct sockaddr *)&raddr, sizeof(raddr)) < 0) {
		printf("Can't connect. %s (%d)\n", strerror(errno), errno);
		close(s);
		return -1;
	}

	context->fd = s;

	return 1;
}


gint btobex_disconnect(obex_t *handle, gpointer userdata)
{
	struct btobex_context_t *context = userdata;
	printf("btobex_disconnect\n");
	close(context->fd);

	return 1;
}


//dli
gint btobex_listen(obex_t *handle, gpointer userdata)
{
	struct btobex_context_t *context = userdata;
	struct sockaddr_rc laddr, raddr;
	int s, rs;
  int alen ;
  	static int connected_sk = -1;

	printf("btobex_listen\n");
	if (connected_sk > 0) close( connected_sk); //le socket a t ouvert avant et n'est toujours pas ferm. todo:  debugger
 	connected_sk = -1;

  if ((s = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
		printf("Can't create socket. %s (%d)\n", strerror(errno), errno);
		return -1;
	}

	laddr.rc_family = AF_BLUETOOTH;
	bacpy(&laddr.rc_bdaddr, BDADDR_ANY);
	laddr.rc_channel = context->channel;
	printf("bind\n");
	if (bind(s, (struct sockaddr *)&laddr, sizeof(laddr)) < 0) {
		printf("Can't bind socket. %s (%d)\n", strerror(errno), errno);
		close(s);
		exit(1);
	}

  printf("listen\n");
  listen(s, 10) ;

// 	obex_push_sdp_register(context->channel) ;
  
	raddr.rc_family = AF_BLUETOOTH;
	raddr.rc_channel = context->channel;
  alen = sizeof(raddr) ;
  rs = accept(s, (struct sockaddr *)&raddr, &alen) ;
  if (rs < 0)
  {
    printf("Can't accept socket. %s (%d)\n", strerror(errno), errno) ;
    close(rs) ;
    close(s) ;
    return -1 ;
  }
  else
  {
		char str[40];
		
    ba2str(&raddr.rc_bdaddr, str);
		printf("New connection from %s", str);
		bacpy(&context->bdaddr, &raddr.rc_bdaddr);
  }

	close(s) ;
  connected_sk = context->fd = rs;

	return 1;
}


gint btobex_write(obex_t *handle, gpointer userdata, guint8 *buffer, gint length)
{
	struct btobex_context_t *context = userdata;
	return (write(context->fd, buffer, length));
}


gint btobex_handleinput(obex_t *handle, gpointer userdata, gint timeout)
{
	struct btobex_context_t *context = userdata;
	struct timeval to;
	fd_set fdset;
	unsigned char buf[1024];
	int sel, len;

  to.tv_sec = timeout;
	to.tv_usec = 0;

	FD_ZERO(&fdset);
	FD_SET(context->fd, &fdset);

	len = -1;

	if ((sel = select(context->fd + 1, &fdset, NULL, NULL, &to)) > 0) {
		if ((len = read(context->fd, buf, sizeof(buf))) > 0)
    {
    	OBEX_CustomDataFeed(handle, buf, len);
    }
	}

	return len;
}


obex_ctrans_t btobex_ctrans = {
	customdata:    &btobex_context,
	connect:     btobex_connect,
	disconnect:  btobex_disconnect,
	listen:      btobex_listen,
	write:       btobex_write,
	handleinput: btobex_handleinput,
};


void btobex_event(obex_t *handle, obex_object_t *object, gint mode, gint event, gint obex_cmd, gint obex_rsp)
{
	struct btobex_context_t *srv ;
  int ret ;
  

  srv = OBEX_GetUserData(handle) ;
  
 	switch (event) {
	case OBEX_EV_PROGRESS:
		break;
	case OBEX_EV_ABORT:
		printf("Request aborted\n");
		finished = 1;
		break;
//	case OBEX_EV_REQDONE:
//		finished = 1;
//		break;
//	case OBEX_EV_REQHINT:
//		printf("Request hint\n");
//		break;
// handeled by server
//	case OBEX_EV_REQ:
//		printf("Server request\n");
		break;
	case OBEX_EV_LINKERR:
		OBEX_TransportDisconnect(handle);
		printf("Link broken\n");
		break;
	case OBEX_EV_PARSEERR:
		printf("Parse error\n");
		break;
//dli additional events supported when server
	case OBEX_EV_STREAMAVAIL:
		ret = btobex_srv_receive(srv, object, FALSE);
		break;
	case OBEX_EV_REQ:
		printf("Incoming request %02x, %d\n", obex_cmd, OBEX_CMD_SETPATH);

		switch(obex_cmd) {
		case OBEX_CMD_CONNECT:
			printf("Connected\n");	
			srv->obexhandle = handle ;
      srv->input_file = -1 ;
			srv->fifo_pump_pid = -1;
      ret = 1;
			break;
		case OBEX_CMD_DISCONNECT:
			printf("OBEX_CMD_DISCONNECT\n");
			ret = 1;
			break;

		case OBEX_CMD_PUT:
      printf("Receive\n") ;
			ret = btobex_srv_receive(srv, object, TRUE);
			break;

		case OBEX_CMD_SETPATH:
      printf("Set path\n") ;
			ret = btobex_srv_setpath(srv, object);
			break;
		default:
			ret = 1;
			break;
		}

		if(ret < 0) {
			finished = TRUE;
			success = FALSE;
		}
			
		break;

	case OBEX_EV_REQHINT:
		/* An incoming request is about to come. Accept it! */
		switch(obex_cmd) {
		case OBEX_CMD_PUT:
			printf("Going to turn streaming on!\n");
			/* Set response to ok! */
			OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
			/* Turn streaming on */
      printf("Reading stream\n") ;
			OBEX_ObjectReadStream(handle, object, NULL);
			break;
		
		case OBEX_CMD_SETPATH:
		case OBEX_CMD_CONNECT:
		case OBEX_CMD_DISCONNECT:
			/* Set response to ok! */
			OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
			break;
		
		default:
			printf("Skipping unsupported command:%02x\n", obex_cmd);
			OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
			break;
		
		}
		break;

	case OBEX_EV_REQDONE:
		if(obex_cmd == OBEX_CMD_DISCONNECT) {
			finished = TRUE;
			success = TRUE;
		}
		break;
    
  default:
		printf("Unknown event 0x%02x\n", event);
		break;
	}
}

static gint new_file(struct btobex_context_t *srv, obex_object_t *object)
{
	obex_headerdata_t hv;
	guint8 hi;
	gint hlen;
	gchar *name = NULL;
	gint ret = -1;
	//char full_name[128] ;
	int file_length = -1;
	char file_length_str[10];
  
	/* First iterate through recieved header to find name */
	while(OBEX_ObjectGetNextHeader(srv->obexhandle, object, &hi, &hv, &hlen))	{
		switch(hi)	{
		case OBEX_HDR_NAME:
			if( (name = /*g_malloc*/malloc(hlen / 2)))	{
				OBEX_UnicodeToChar(name, hv.bs, hlen);
			}
			break;
			
		case OBEX_HDR_LENGTH:
			printf("total length: %d\n", hv.bq4 );
			file_length = hv.bq4;
			break;
			
		default:
			printf("Skipped header %02x\n", hi);
		}
	}
	if(name == NULL)	{
		printf("Got a PUT without a name. Refusing\n");
		/* Send back error */
		OBEX_ObjectSetRsp(object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
		printf("Error\n");
		goto out;
	}

	printf("Receiving %s\n", name);
	strncpy(srv->name, name, 128) ;

	//on lance le script qui va attendre sur la fifo
	char str[40];
	ba2str(&(srv->bdaddr), str);
	printf("Start obex_ftp script: %s - args: %s %s %s %d\n", inbox_script, srv->fifo_path, srv->name, str, file_length);

	pid_t pid;
	if ((pid = fork ()) < 0)
	{
		//can't fork
		ret = -1;
		goto out;
	}

	if (pid == 0)
	{
		int i;
		printf("start execle\n");

		/* close every file descriptor other than 0,1,2 to make things clean*/
		for (i=3; i < 64; i++) close (i);

		sprintf( file_length_str, "%d", file_length );
		execl( inbox_script,inbox_script, srv->fifo_path, srv->name, str ,file_length_str, NULL);
		perror("impossible d'executer le script " );
		exit(127);
	}
	srv->fifo_pump_pid = pid;
	printf("%s est maintenant lance\n",inbox_script);


	printf("transfert to fifo ...%s\n", srv->fifo_path) ;
  
	srv->input_file = open(srv->fifo_path, O_CREAT | O_TRUNC | O_WRONLY , 0);
	ret = srv->input_file;

	printf("input_file: %d  (context: %p)\n", ret, srv);

	if (ret < 0)
	{
		printf("can't open fifo %s\n", srv->fifo_path);
		goto out;
	}


out:	/*g_free*/free(name);
	return ret;
}

gint btobex_srv_setpath(struct btobex_context_t *srv, obex_object_t *object)
{
  printf("Don't care about path\n") ;
	return 0 ;
}
//
// Extract interesting things from object and save to disk.
//
gint btobex_srv_receive(struct btobex_context_t *srv, obex_object_t *object, gboolean finished)
{
	const guint8 *body = NULL;
	gint body_len = 0;
	int status;

//  char cmd[256] ;

	if(srv->input_file < 0 && finished == FALSE) {
		/* Not receiving a file */
    printf("Not receiving a file\n") ;
		if(new_file(srv, object) < 0)
			return 1;
	}
	
	if(finished == TRUE) {
        	/* Recieve done! */
		printf("Done!...\n");
		if (srv->input_file > 0)
		{
			printf("close file fd %d\n",srv->input_file);
			close(srv->input_file);
			srv->input_file = -1;
			//wait children exit and test status
			printf("wait fifo pump exit (pid %d)\n", srv->fifo_pump_pid);
			waitpid( srv->fifo_pump_pid, &status, 0 );
			printf("  wait_pid done with status %d\n", status );
			srv->fifo_pump_pid = -1;
			if (WIFEXITED(status) && ( WEXITSTATUS( status ) != 0)) {
				//chidren exit with error
				printf("Child exit with error\n");
				OBEX_ObjectSetRsp(object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
				return -1;
			}
		}
		return 1;
	}
	else if (srv->input_file > 0) {
		/* fd is valid. We are currently receiving a file */
		body_len = OBEX_ObjectReadStream(srv->obexhandle, object, &body);
		
		if(body_len < 0) {
			/* Error */
		}
		else if(body_len == 0) {
   		//char str[40];
			/* EOS */
			printf("EOF (close fd %d)\n",srv->input_file);
			close(srv->input_file);
			srv->input_file = -1;
			//wait children exit and test status
			printf("wait fifo pump exit (pid %d)\n", srv->fifo_pump_pid);
			waitpid( srv->fifo_pump_pid, &status, 0 );
			printf("  wait_pid done with status %d\n", status );
			srv->fifo_pump_pid = -1;
			if (WIFEXITED(status) && ( WEXITSTATUS( status ) != 0)) {
				//chidren exit with error
				printf("Child exit with error\n");
				OBEX_ObjectSetRsp(object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
				return -1;
			}

/*
	    ba2str(&(srv->bdaddr), str);

      sprintf(cmd, "%s \"%s\" \"%s\" %s &",
          inbox_script, srv->full_name, srv->name, str) ;
      printf("Executing %s\n", cmd) ;
      system(cmd) ;
     */
		}
		else {
			if(srv->input_file > 0)
      {
				write(srv->input_file, body, body_len);
      }
		}
		return 1;
	}
	return -1;
}	


guint8 *btobex_readfile(const char *filename, int *filesize)
{
	guint8 *buf;
	struct stat stats;
	int fd, fs;

	stat(filename, &stats);
	fs = stats.st_size;

	if ((fd = open(filename, O_RDONLY, 0)) < 0)
		return NULL;

	if (!(buf = /*g_malloc*/malloc(fs))) {
		close(fd);
		return NULL;
	}

	*filesize = read(fd, buf, fs);

	close(fd);

	return buf;
}

int btobex_pull(char *fifo_path, uint8_t channel)
{
	obex_t *handle;
	int err;

	if (!(handle = OBEX_Init(OBEX_TRANS_CUST, btobex_event, 0))) {
		printf("Init of OBEX failed. %s (%d)", strerror(errno), errno);
		return -1;
	}

	btobex_context.channel = channel;
	btobex_context.fd = -1;
	btobex_context.fifo_pump_pid = -1;
	strncpy(btobex_context.fifo_path, fifo_path, 128) ;
  
	OBEX_SetUserData(handle, &btobex_context);

	if (OBEX_RegisterCTransport(handle, &btobex_ctrans) < 0) {
		printf("Custom transport callback registration failed.");
		return -1;
	}

	if (OBEX_ServerRegister(handle, (void *)1, 0) != 1)
		return -1;

	finished = 0;
	while (!finished) {
		if ((err = OBEX_HandleInput(handle, 1)) < 0) {
			printf("btobex_pull: Error while handling input.\n");
			break;
		}
	}

	/* cleanup */
	printf("cleanup...(context %p)\n", &btobex_context);
	if (btobex_context.input_file >= 0) {
		printf("cleanup fifo fd %d\n", btobex_context.input_file);
		close( btobex_context.input_file );
		btobex_context.input_file = -1;
	}

	if (btobex_context.fifo_pump_pid > 0) {
		printf("cleanup fifo pump (pid %d)\n", btobex_context.fifo_pump_pid);
		//kill( btobex_context.fifo_pump_pid, SIGTERM );
		printf("wait fifo pump exit (pid %d)\n", btobex_context.fifo_pump_pid);
		waitpid( btobex_context.fifo_pump_pid, &err, 0 );
		printf("  wait_pid done with status %d\n", err );
		btobex_context.fifo_pump_pid = -1;
	}
	return 0;
}

int btobex_push(bdaddr_t *bdaddr, uint8_t channel, char *filename, char *alias)
{
	obex_t *handle;
	obex_object_t *object;
	obex_headerdata_t hd;

	guint8 namebuf[MAXPATHLEN + 1];
	int namelen;
	guint8 *databuf;
	int datalen;

	int err;

	if (!(handle = OBEX_Init(OBEX_TRANS_CUST, btobex_event, 0))) {
		printf("Init of OBEX failed. %s (%d)", strerror(errno), errno);
		return -1;
	}

	bacpy(&btobex_context.bdaddr, bdaddr);
	btobex_context.channel = channel;

	OBEX_SetUserData(handle, &btobex_context);

	if (OBEX_RegisterCTransport(handle, &btobex_ctrans) < 0) {
		printf("Custom transport callback registration failed.");
		return -1;
	}

	if (OBEX_TransportConnect(handle, (void *)1, 0) != 1)
		return -1;


  printf("Connecting...\n") ;
	object = OBEX_ObjectNew(handle, OBEX_CMD_CONNECT);
	err = OBEX_Request(handle, object);

/*
	finished = 0;
	while (!finished) {
		if ((err = OBEX_HandleInput(handle, 1)) < 0) {
			printf("Error while handling input.\n");
			break;
		}
	}
*/
  err = OBEX_HandleInput(handle, 1) ;
	printf("Result : %d\n", err) ;
  object = OBEX_ObjectNew(handle, OBEX_CMD_PUT);

	namelen = OBEX_CharToUnicode(namebuf, alias, MAXPATHLEN);
	hd.bs = namebuf;
  OBEX_ObjectAddHeader(handle, object, OBEX_HDR_NAME, hd, namelen, 0);

  if (!(databuf = btobex_readfile(filename, &datalen))) {
		OBEX_TransportDisconnect(handle);
		return -1;
	}

  hd.bq4 = datalen;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_LENGTH, hd, sizeof(guint32), 0);

  hd.bs = databuf;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY, hd, datalen, 0);

	printf("Sending object ...\n");

	err = OBEX_Request(handle, object);

/*
	finished = 0;
	while (!finished) {
		if ((err = OBEX_HandleInput(handle, 1)) < 0) {
			printf("Error while handling input.\n");
			break;
		}
	}
*/
    
  err = OBEX_HandleInput(handle, 1) ;
	printf("Result : %d\n", err) ;
  printf("Disconnecting\n") ;

	object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT);
	err = OBEX_Request(handle, object);

/*
	finished = 0;
	while (!finished) {
		if ((err = OBEX_HandleInput(handle, 1)) < 0) {
			printf("Error while handling input.\n");
			break;
		}
	}
*/
    
  err = OBEX_HandleInput(handle, 1) ;
	printf("Result : %d\n", err) ;

	OBEX_TransportDisconnect(handle);

	return 0;
}



static void usage(void);


void cmd_push(bdaddr_t *local, int argc, char **argv)
{
	char *filename;
	char *alias;
	bdaddr_t bdaddr;
	uint8_t channel;

	if (argc < 3) {
		usage();
		return;
	}

	filename = argv[1];
	alias = basename(filename);
	str2ba(argv[2], &bdaddr);
	channel = (argc > 3) ? atoi(argv[3]) : 10;

	btobex_push(&bdaddr, channel, filename, alias);
}

void cmd_pull(bdaddr_t *local, int argc, char **argv)
{
	uint8_t channel;
  char *fifo_path ;


	if (argc < 4) {
		usage();
		return;
	}

  	fifo_path= argv[1] ;
	channel = atoi(argv[2]) ;
	strcpy( (char *)inbox_script, argv[3]);

	while(1)
    btobex_pull(fifo_path, channel);
}


struct {
	char *cmd;
	void (*func)(bdaddr_t *bdaddr, int argc, char **argv);
	char *opt;
	char *doc;
} command[] = {
	{ "push", cmd_push, "<file> <bdaddr> [channel]", "Push a file" },
  { "pull", cmd_pull, "<dir> <channel> <script>", "Get a file" },
  { NULL, NULL, 0, 0 }
};


static void usage(void)
{
	int i;

	printf("Bluetooth OBEX tool (" __DATE__ "  " __TIME__ "\n\n");

	printf("Usage:\n"
		"\tobextool [options] <command>\n"
                "\n");

	printf("Options:\n"
		"\t-i [hciX|bdaddr]   Local HCI device or BD Address\n"
		"\t-h, --help         Display help\n"
		"\n");

	printf("Commands:\n");
        for (i = 0; command[i].cmd; i++)
		printf("\t%-6s %-18s\t%s\n", command[i].cmd,
		command[i].opt ? command[i].opt : " ",
		command[i].doc);
	printf("\n");
}


static struct option main_options[] = {
	{ "help",	0, 0, 'h' },
	{ "device",	1, 0, 'i' },
	{ 0, 0, 0, 0 }
};


int main(int argc, char *argv[])
{
	bdaddr_t bdaddr;
	int i, opt;

	bacpy(&bdaddr, BDADDR_ANY);

	while ((opt = getopt_long(argc, argv, "+i:h", main_options, NULL)) != -1) {
		switch(opt) {
		case 'i':
			if (strncmp(optarg, "hci", 3) == 0)
				hci_devba(atoi(optarg + 3), &bdaddr);
			else
				str2ba(optarg, &bdaddr);
			break;
		case 'h':
			usage();
			exit(0);
		default:
			exit(0);
		}
	}

	argc -= optind;
	argv += optind;
	optind = 0;

	if (argc < 1) {
		usage();
		exit(0);
	}

	for (i = 0; command[i].cmd; i++) {
		if (strncmp(command[i].cmd, argv[0], 3))
			continue;
		command[i].func(&bdaddr, argc, argv);
		exit(0);
	}

	usage();

	return 0;
}
