/*
 * Project: ReAim
 *
 * Release: 0.5 alpha
 *
 * Homepage:  http://reaim.sourceforge.net/
 *
 * Copyright: Mark Cooke, November 2001.  All Rights Reserved.
 * 
 * License: GNU General Public License, Version 2.
 *
 * CVS Data: $Id: reaim.c,v 1.1 2004/02/10 11:13:59 cmo Exp $
 *
 * Description:
 *
 * ReAIM is a simple transparent AIM proxy.  It sits between a local area
 * network, and the internet, usually running on the firewall / NAT server.
 *
 * It also has a filter module for MSN Messenger file transfer support.
 *
 * It accepts transparently redirected port 5190 (AIM server communications)
 * traffic, and transparently forwards this to the usual destination with one
 * caveat.
 *
 * The direct-communications setup packet is mangled to use the (routable)
 * internet-visible IP address, rather than the client's NAT / private network
 * address.  This allows the second part of this proxy - a port 4443/5566
 * forwarding service - to catch the remote party DCC connection, and forward
 * it to the internal host.  It is slightly cludgy, as the AIM clients do not
 * send a 'hello, it's me' packet when they direct-connect, so the proxy can't
 * tell where to forward the connection to until the remote party sends a
 * message.  Fortunately, just starting to type sends a message we can use
 * to make the onward part of the proxied connection.
 *
 * As a compile option, setting LAX_SECURITY will allow a DCC to match at the
 * initial connection point if there is only 1 pending DCC request.
 * 
 * As a compile option, setting LOOSE_CLIENT_IP will disable the paranoid
 * checking that the IP address being overwritten in the DCC setup packet is
 * the same as the source IP address of the client socket.  This allows
 * multi-homed hosts to work, and also network topologies that include
 * multiple-NAT devices.
 *
 * Usage:
 *
 * reaim [optional dotted quad for routable IP]
 *
 * Notes:
 *
 * This proxy requires knowledge of the external IP address, and also for the
 * original destination to be available.  Using iptables allows us to use
 * SO_ORIGINAL_DST to retrieve this.  Hence, this proxy is a 2.4.x / iptables
 * only proxy.  More specifically, it does -not- work with 2.4.x / ipchains
 * support.  Suitable modifications to the gettranssockname function may
 * permit this to work with ipchains on 2.2.x, where the getsockname() call
 * is overridden to be the original destination (not confirmed).
 *
 * Example usage where eth0 is LAN, ppp0 is internet:
 *
 * 1. Redirect eth0 traffic to eth0 port 5190, and make sure it's accepted.
 * iptables -t nat    -A PREROUTING -i eth0 -p tcp --dport 5190 -j REDIRECT --to-ports 5190
 * iptables -t filter -A INPUT      -i eth0 -p tcp --dport 5190 -j ACCEPT
 *
 * 2. Allow ppp0 to accept traffic for port 4443 (AIM DCC)
 * iptables -t filter -A INPUT      -i ppp0 -p tcp --dport 4443 -j ACCEPT
 *
 * 3. Allow ppp0 to accept traffic for port 5566 (AIM Sharing)
 * iptables -t filter -A INPUT      -i ppp0 -p tcp --dport 5566 -j ACCEPT
 *
 * 4. Redirect eth0 traffic to eth0 port 5190, and make sure it's accepted.
 * iptables -t nat    -A PREROUTING -i eth0 -p tcp --dport 1863 -j REDIRECT --to-ports 1863
 * iptables -t filter -A INPUT      -i eth0 -p tcp --dport 1863 -j ACCEPT
 *
 * 5. Allow ppp0 to accept traffic for port 1864 (MSN DCC)
 * iptables -t filter -A INPUT      -i ppp0 -p tcp --dport 1864 -j ACCEPT
 *
 *
 *
 * Detailed AIM Manglement Notes:
 *
 * The proxy keeps three lists of connections (socket pairs).
 *
 * 1. LAN clients connections to port 5190 (and the onward connection).
 * 2. Established connections on port 4443 (and the onward connection).
 * 3. Established connections on port 5566 (and the onward connection).
 *
 * When a client starts a connection, the proxy accepts the redirected
 * connection on it's port 5190.  It immediately attempts to open the
 * onward connection to the original destination, and any packets arriving
 * in either direction are inspected, optionally modified, and then passed
 * to the other half of the socket-pair.
 *
 * When a packet indicates a DC setup request, an entry is made in the
 * pending DC setup list, and appropriate parameters from the request are
 * recorded.  Note that the DC setup request is -mangled- to use the
 * externally addressable IP address, so the remote party isn't trying to
 * connect to a 10.x.x.x etc IP address.
 *
 * When a connection is received on port 4443, the proxy accepts the
 * connection, and searches for a match in the pending DC list.  If a
 * match is found, the onward portion of the DC is established, and 
 * unfiltered direct communications between the parties is established.
 *
 * Connections to port 5566 are handled virtually identically. (Indeed
 * the current code is just a cut and paste job, with small mods)
 *
 * DC Setup Packet Format:
 *
 * With many thanks to the people working on the reverse engineering of
 * the AIM OSCAR protocol.
 *
 *
 * EXTRA...EXTRA...
 *
 * Because of the nice flexibility of the filter scheme, I added support for
 * MSN proxying too.  The MSN Messenger uses a pseudo HTTP style scheme, with
 * headers indicating the message type, parameters, etc.
 *
 * For MSN DCC setup, we look for the text/xmsmsgsinvite message, and 'fix'
 * both the port and IP address.
 *
 * TODO:
 *
 * A whole lot.  This is currently little more than a complete hack that
 * happens to work for me.  I'll be working on the code to do more error
 * checking, structure clean up, as time permits.
 *
 * I have plans to add a 'filter chain' type system to both the 5190 and 4443
 * port proxies, and eventually maybe generalise the code to be a more
 * generic, open source licensed, proxy-based firewalling system...maybe!
 *
 * As an option, it would be possible to listen on a range of ports for direct
 * connection establishment - and this is a possible way to work around the
 * ambiguous nature of having a single proxy listen port for multiple machines.
 * This needs thinking about and some testing.
 *
 * Final words:
 *
 * No warranty.  Patches, bug reports, success stories, and other contributions
 * most welcome.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <sys/time.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <time.h>
#include <ctype.h>

#include <getopt.h>
#include <sys/ioctl.h>

#include <unistd.h>


// Nasty kernel include in a userspace program
// but we need the original destination information....
#include <linux/netfilter_ipv4.h>

#define REAIM_VERSION "0.5 alpha"

#define OSCAR 5190
#define DCC   4443
#define SHARE 5566

#define MSN   1863   // Where we listen for client messages
#define MSNRX 1864   // Where we listen for MSN DCCs

#define ICQ_MIN 30000 // Minimum port for icq direct connections
#define ICQ_MAX 30099 // Maximum port for icq direct connections

#define LAX_SECURITY     1
#define LOOSE_CLIENT_IP  1
#define LOOSE_PORT_SCAN  0

#define TS_CLIENT 0
#define TS_SERVER 1

// This defines the size of the buffers used when proxying
// connections.  Each proxied connection uses 2 of these
// buffers.
#define PROXY_BUFFER_SIZE   4096

#define LOG_FILE "/var/log/reaim.log"

// Filter group IDs - needed for stateful storage access
#define GRP_AIM    1
#define GRP_ICQ    2
#define GRP_MSN    3

// Logging routine.  Borrowed from the squirm proxy.
int write_log(const char *logfile, const char *format, ...);


typedef struct _filter_chain         filter_chain;
typedef struct _filter_state         filter_state;
typedef struct _proxy_track_struct   proxy_track_struct;
typedef struct _user_track_struct    user_track_struct;
typedef struct _pending_dc_struct    pending_dc_struct;
typedef struct _listen_track_struct  listen_track_struct;
typedef struct _pending_msn_struct   pending_msn_struct;


/*----------------------------------------------------------------------------
  The filter scheme for the code is a simple procedure that takes a proxy
  data buffer, current and maximum length, and the connection tracking
  structure.

  Each filter is called sequentially, and can mangle the provided buffer.
----------------------------------------------------------------------------*/
typedef unsigned int (filter_proc)(unsigned char *buff,
				   unsigned int buff_len,
				   unsigned int max_len,
				   proxy_track_struct *pt);

typedef void (listen_proc)(listen_track_struct *lt);


struct _filter_chain {
    filter_chain   *next;    // Next in filter chain
    filter_proc    *proc;    // Procedure for filtering
};

struct _filter_state {
    filter_state   *next;         // Next state storage
    unsigned int    filter_group; // Filter ID number
    void           *group_state;  // Persistent state
};

/*
 * Generic proxied socket handler.
 *
 */
struct _proxy_track_struct {
    proxy_track_struct *next;  // Next proxy connection in linked list.
    
    unsigned int magic;        // Magic number linking other lists.
    unsigned int cleanup;      // Set to 1 to garbage collect this entry.
    unsigned int user;         // Set to 1 if user connection.
    
    int sockets[2];                  // The rx / tx sockets
    unsigned char *data_buff[2];     // Receive/Transmit pending data buffers
    unsigned int   data_offset[2];   // Offset to start of data in buffers
    unsigned int   data_len[2];      // Remaining bytes in pending buffers
    unsigned int   data_max[2];      // Buffer sizes
    filter_chain  *filter[2];        // Attached filters
    filter_state   state;            // Stateful storage buffers
};

struct  _user_track_struct {
    user_track_struct *next; // Next entry in the list

    unsigned int magic;      // Magic connection tracking number
    unsigned int cleanup;    // Set to 1 to clean up this entry

    unsigned char *msg;      // Connection setup message
    unsigned char *user;     // Screen name of the user
};

struct _pending_dc_struct {

    pending_dc_struct *next; // Next pending connection
    
    unsigned int cleanup;    // Set to 1 to clean up this entry

    unsigned short port;     // Port we're listenning on.

    unsigned int magic;      // The user_track_struct magic number
    unsigned char id[8];     // The DC setup message we're waiting on

};

struct _pending_msn_struct {

    pending_msn_struct *next; // Next pending connection
    
    unsigned int cleanup;    // Set to 1 to clean up this entry

    unsigned char ip[4];     // Client IP address
    int client_port;         // Port the client is listenning on
    unsigned int cookie;     // Handshake token
};

struct _listen_track_struct {
    
    listen_track_struct *next; // Next listenning socket
    listen_proc         *proc; // Procedure to call
    
    unsigned int      cleanup; // 1 if inactive
    int                  port; // Port this is listenning on.
    int                socket; // Socket handle.
};

#define AIM_CMDSTART(buff)      buff[0]
#define AIM_CHANNEL(buff)       buff[1]
#define AIM_SEQID(buff)         ((unsigned short) buff[2]  + (buff[3] << 8))
#define AIM_FNAC_LEN(buff)      ((unsigned short) buff[5]  + (buff[4] << 8))
#define AIM_FNAC_FAMILY(buff)   ((unsigned short) buff[7]  + (buff[6] << 8))
#define AIM_FNAC_SUBTYPE(buff)  ((unsigned short) buff[9]  + (buff[8] << 8))
#define AIM_FNAC_FLAGS(buff)    ((unsigned short) buff[11] + (buff[10] << 8))
#define AIM_SNAC_ID(buff)       ((unsigned int)   buff[15] + (buff[14] << 8) + (buff[13] << 16) + (buff[12] << 24))
#define AIM_SNAC_COOKIE(buff)   (&buff[16])   // The 8 byte SNAC cookie start ptr
#define AIM_SNAC_CHAN(buff)     ((unsigned short) buff[25] + 256*buff[24])
#define AIM_BUD_NAME_LEN(buff)  buff[26]
#define AIM_BUD_NAME(buff)      (&buff[27])
#define AIM_BUD_NEXT(buff)      (27+AIM_BUD_NAME_LEN(buff))
#define AIM_SNAC_TLV_TYPE(buff) ((unsigned short) buff[AIM_BUD_NEXT(buff)+1] + (buff[AIM_BUD_NEXT(buff)+0] << 8))
#define AIM_SNAC_TLV_NEXT(buff) ((unsigned short) buff[AIM_BUD_NEXT(buff)+3] + (buff[AIM_BUD_NEXT(buff)+2] << 8))
#define AIM_SNAC_DATA_OFF(buff) (AIM_BUD_NEXT(buff)+4)
#define DCC_SNAC_COOKIE(buff)   (&buff[12])
#define DCC_BUD_NAME(buff)      (&buff[44])  // Appears to be null terminated this time...

/*===========================================================================*/
// The linked lists used to track state
/*===========================================================================*/

proxy_track_struct pt_list;       // Generic proxy socket list-head.
user_track_struct  user_list;     // List head.  User tracking information
listen_track_struct lt_list;      // List head.  Listenning sockets

pending_dc_struct *pending_dc_list = NULL; // This pair are 4443 port proxies

pending_msn_struct *pending_msn_list = NULL; // This pair are MSN XFer proxies

unsigned int       ts_magic        = 1; // Next free magic number

char *             outside_ip      = NULL;

filter_chain       oscar_tx;   // LAN client to AIM filter
filter_chain       oscar_rx;   // AIM to LAN client filter
filter_chain       dcc_tx;     // LAN to Remote filter
filter_chain       dcc_rx;     // Remote to LAN filter
filter_chain       share_tx;   // LAN to Remote filter
filter_chain       share_rx;   // Remote to LAN filter
filter_chain       msn_tx;     // LAN to Remote filter
filter_chain       msn_rx;     // Remote to LAN filter
filter_chain       msnrx_tx;   // LAN to Remote filter
filter_chain       msnrx_rx;   // Remote to LAN filter

/*===========================================================================*/


/*===========================================================================*
 *  int gettranssockname(int fd, struct sockaddr *sa, socklen_t *salen)
 *
 *    Query the network stack for the original destination of the socket
 *    passed.  This is used for building the onward portion of the filter
 *    proxy.
 *===========================================================================*/

int gettranssockname(int fd, struct sockaddr *sa, socklen_t *salen)
{
    if (*salen != sizeof(struct sockaddr_in)) {
	errno = EINVAL;
	return -1;
    }
    
    // Try to ask for the original destination...
    if (!getsockopt(fd, SOL_IP, SO_ORIGINAL_DST, sa, salen))
	return 0;
    
    // We have to explode if we can't read it back.
    return -1;
}

/*===========================================================================*
 *  void dump_packet(const char *buff, int len)
 *
 *    Print out the contents of the packet.
 *===========================================================================*/
void dump_packet(const char *buff, int len)
{
    char dump_start[32768], *dump_buff = dump_start;
    int i, j;

    for (i = 0; i < len; i ++) {
	
	if ((i % 16) == 0) {
	    sprintf(dump_buff, "%.4X    ", i);
	    dump_buff += 8;
	}
	
	sprintf(dump_buff, "%02x ", buff[i]);
	dump_buff += 3;
	
	if ((i % 16) == 15) {
	    
	    sprintf(dump_buff, "    ");
	    dump_buff += 4;
	    
	    for (j = -15; j < 1; j++) {
		if (isalnum(buff[i+j]))
		    *dump_buff = buff[i+j];
		else
		    *dump_buff = '.';

		dump_buff ++;
	    }
	    
	    *dump_buff++ = '\n';
	}
    }
    *dump_buff++ = '\n';
    *dump_buff++ = '\0';

    write_log(LOG_FILE, "%s", dump_start);
}


/*===========================================================================*
 *  void dump_packet(const char *buff, int len)
 *
 *    Print out the contents of the packet.
 *===========================================================================*/

int write_log(const char *logfile, const char *format, ...)
{
    va_list args;
    FILE *log_handle;
    char *date_str;
    char msg_buff[32768];
    time_t current_time;
    
    va_start(args, format);
    
    if (vsprintf(msg_buff, format, args) > 32767)
	strcpy(msg_buff, "[WARNING] Insufficient log buffer size");
    
    va_end(args);
    
    current_time = time(NULL);
    date_str = ctime(&current_time);
    
    log_handle = fopen(logfile, "at");
    if (log_handle == NULL)
	return 0;
    
    // strlen - 1 to kill the trailing \n in the data string.
    fprintf(log_handle, "[%.*s] %s\n", (int) strlen(date_str) - 1, date_str, msg_buff);
    fclose(log_handle);

    return 1;
}

/*===========================================================================*
 *  int net_listen(int port)
 *
 *    Open a network socket listenning on the port requested.  This function
 *    returns the socket handle, or -1 if the socket could not be openned for
 *    some reason.
 *===========================================================================*/

int net_listen(int port, const char *ifs)
{
    int hSock;
    struct sockaddr_in server;
    int option;
    
    hSock = socket(PF_INET, SOCK_STREAM, 0);
    if (hSock == -1) {
	perror("FATAL: Could not create socket");
	return -1;
    }
    
    option = 1;
    if (setsockopt(hSock, SOL_SOCKET, SO_REUSEADDR, (void*) &option,
		   sizeof(option)) < 0) {
	perror("FATAL: Could not set socket options");
	return -1;
    }
    
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = inet_addr(ifs);
    server.sin_port = htons(port);
    if (bind(hSock, (struct sockaddr*) &server, sizeof(server)) == -1) {
	perror("FATAL: Could not bind socket");
	close(hSock);
	return -1;
    }
    
    if (listen(hSock, 5) == -1) {
	perror("FATAL: Could not listen");
	close(hSock);
	return -1;
    }
    
    /* Set the socket to non blocking, and close it if we can't */
    if (fcntl (hSock, F_SETFL, O_NONBLOCK) == -1) {
	perror("FATAL: Could not setup non-blocking");
	close(hSock);
	return -1;
    }
    
    return hSock;
}

/*===========================================================================*
 *  int net_connect_to(int port, const char *ifs)
 *
 *    Open a network socket to the port and dotted-quad address string given.
 *    The returned socket is set to non-blocking and keep-alive if the socket
 *    was created.  -1 is returned on failure.
 *===========================================================================*/

int net_connect_to(int port, const char *ifs)
{
    int hSock;
    struct sockaddr_in server;
    int option;
    
    hSock = socket(PF_INET, SOCK_STREAM, 0);
    if (hSock == -1) {
	write_log(LOG_FILE, "Connection [%s:%d] failed [create]: %s [%d]",
		  ifs, port, sys_errlist[errno], errno);
	return -1;
    }
    
    option = 1;
    if (setsockopt(hSock, SOL_SOCKET, SO_REUSEADDR, (void*) &option,
		   sizeof(option)) < 0) {
	write_log(LOG_FILE, "Connection [%s:%d] failed [options]: %s [%d]",
		  ifs, port, sys_errlist[errno], errno);
	close(hSock);
	return -1;
    }
    
    option = 1;
    if (setsockopt(hSock, SOL_SOCKET, SO_KEEPALIVE, (void*) &option,
		   sizeof(option)) < 0) {
	write_log(LOG_FILE, "Connection [%s:%d] failed [options]: %s [%d]",
		  ifs, port, sys_errlist[errno], errno);
	close(hSock);
	return -1;
    }
    
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = inet_addr(ifs);
    server.sin_port = htons(port);
    if (connect(hSock, (struct sockaddr*) &server, sizeof(server)) == -1) {
	write_log(LOG_FILE, "Connection [%s:%d] failed [conn]: %s [%d]",
		  ifs, port, sys_errlist[errno], errno);
	close(hSock);
	return -1;
    }
    
    /* Set the socket to non blocking, and close it if we can't */
    if (fcntl (hSock, F_SETFL, O_NONBLOCK) == -1) {
	write_log(LOG_FILE, "Connection [%s:%d] closed [options]: %s [%d]",
		  ifs, port, sys_errlist[errno], errno);
	close(hSock);
	return -1;
    }
    
    return hSock;
}

/*-------------------------------------------------------------------------
  void* net_open_connect(int hSock)
  
  Open the next connection to the socket specified.  The socket supplied
  must be in a listenning state.  This function returns a newly created
  socket structure with information about the new connection, or -1 if
  the socket could not be openned for some reason.
  
  The returned socket is setup for non-blocking, keep-alive.
-------------------------------------------------------------------------*/
int net_open_connect(int hSock, struct sockaddr_in *addr, int *salen)
{
    int hNewSock;
    int option;
    
    hNewSock = accept(hSock, (struct sockaddr *) addr, salen);
    if (hNewSock == -1) {
	write_log(LOG_FILE, "Accept failed [accept]: %s [%d]",
		  sys_errlist[errno], errno);
	return -1;
    }
    
    /* Set the socket to non blocking, and close it if we can't */
    if (fcntl (hNewSock, F_SETFL, O_NONBLOCK) == -1) {
	write_log(LOG_FILE, "Accept failed [options] [From %s:%d]: %s [%d]",
		  inet_ntoa(addr->sin_addr), ntohs(addr->sin_port),
		  sys_errlist[errno], errno);
	close(hNewSock);
	return -1;
    }
    
    option = 1;
    if (setsockopt(hNewSock, SOL_SOCKET, SO_KEEPALIVE, (void*) &option,
		   sizeof(option)) < 0) {
	write_log(LOG_FILE, "Accept failed [options] [From %s:%d]: %s [%d]",
		  inet_ntoa(addr->sin_addr), ntohs(addr->sin_port),
		  sys_errlist[errno], errno);
	close(hNewSock);
	return -1;
    }
    
    return hNewSock;
}

/*-------------------------------------------------------------------------
  filter_state* find_state(proxy_track_struct *pts,
                           unsigned int  id)
  
  Search the list of filter state structures for the PTS, and return a
  pointer to the appropriate one.

  Return NULL if it does not exist.
-------------------------------------------------------------------------*/

filter_state *find_pts_state(proxy_track_struct *pts, unsigned int group_id)
{
    filter_state *scan;
    
    scan = pts->state.next;
    
    while((scan) && (scan->filter_group != group_id))
	scan = scan->next;
    
    return scan;
}

/*-------------------------------------------------------------------------
  int add_pts_state(proxy_track_struct *pts,
                    unsigned int  id,
	            filter_state *add_state)
  
  Search the list of filter state structures for the PTS, and if a filter
  state structure doesn't already exist, add the provided one.

  Return 0 if the new data was not added.  1 if it was.
-------------------------------------------------------------------------*/

int add_pts_state(proxy_track_struct *pts, unsigned int group_id, filter_state *add_state)
{
    if (find_pts_state(pts, group_id))
	return 0;

    add_state->next = pts->state.next;
    pts->state.next = add_state;

    return 1;
}

/*-------------------------------------------------------------------------
  filter_state* del_pts_state(proxy_track_struct *pts,
                              unsigned int  id)
  
  Search the list of filter state structures for the PTS, and if a filter
  state structure exists, remove it from the list and return it.

  Return NULL if there was no pts data.
-------------------------------------------------------------------------*/

filter_state * del_pts_state(proxy_track_struct *pts, unsigned int group_id)
{
    filter_state *scan, *prev;
    
    prev = &pts->state;
    scan = pts->state.next;
    
    while(scan) {
	
	if (scan->filter_group == group_id) {
	    prev->next = scan->next;
	    return scan;
	}
	
	prev = scan;
	scan = scan->next;
    }
    
    return NULL;
}

/*-------------------------------------------------------------------------
  proxy_track_struct* pts_allocate()
  
  Allocate the buffers, setup counters, sizes, empty filters for a new
  proxy tracking structure.

  Return NULL if we can't allocate memory for some reason.
-------------------------------------------------------------------------*/

proxy_track_struct *pts_allocate()
{
    proxy_track_struct *pt;
    int sock;
    
    // Setup the tracking structure
    pt = malloc(sizeof(proxy_track_struct));
    if (!pt) {
	write_log(LOG_FILE, "[WARNING] Memory allocation failed for "
		  "proxy tracking structure.");
	return NULL;
    }

    memset(pt, 0, sizeof(proxy_track_struct));
    
    pt->next       = NULL;
    pt->cleanup    = 0;
    pt->user       = 0;
    pt->state.next = NULL;
    
    // Setup the new user socket list.
    for (sock = 0; sock < 2; sock ++) {
	pt->sockets[sock]      = -1;
	pt->data_max[sock]     = PROXY_BUFFER_SIZE;
	pt->data_offset[sock]  = 0;
	pt->data_len[sock]     = 0;
	pt->data_buff[sock]    = malloc(pt->data_max[sock]);

	if (!pt->data_buff[sock]) {

	    int fs;
	    
	    for (fs = 0; fs < sock; fs ++)
		free(pt->data_buff[fs]);
	    free(pt);
	    
	    write_log(LOG_FILE, "[WARNING] Memory allocation failed for "
		      "proxy tracking data buffers.");
	    
	    return NULL;
	}

	pt->filter[sock] = NULL;
    }
    
    return pt;
}

/*-------------------------------------------------------------------------
  void* net_open_aim_oscar(listen_track_struct *lt)
  
  Open the connection attempt from the LAN client, create the onward
  socket for the proxying, and allocate a tracking structure.
  
  If any steps fail, the client connection is terminated.
-------------------------------------------------------------------------*/
// Open a new client
void net_open_aim_oscar(listen_track_struct *lt)
{
    int new_client, new_server;
    user_track_struct  *ts;
    proxy_track_struct *pt;
    struct sockaddr_in orig_dst_addr, src_addr;
    int salen;
    char *tmp_str;
    
    // Open the pending connection (and log any errors)
    salen = sizeof(src_addr);
    new_client = net_open_connect(lt->socket, &src_addr, &salen);
    if (new_client == -1)
	return;
    
    // Read the original destination...
    salen = sizeof(orig_dst_addr);
    if (gettranssockname(new_client, (struct sockaddr*) &orig_dst_addr, &salen) == -1) {
	write_log(LOG_FILE, "Connection [%s:%d -> proxy -> ??:??] failed "
		  "[Could not determine original destination]: %s [%d]",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
		  sys_errlist[errno], errno);
	close(new_client);
	return;
    }
    
    // This shouldn't ever get tripped, but just incase someone futzed the
    // redirect rules, we check anyway.
    //
    // Added an override to allow source ports != OSCAR.  Not recommended.
    //
#if LOOSE_PORT_SCAN
    if (ntohs(orig_dst_addr.sin_port) != OSCAR) {
	
	tmp_str = strdup(inet_ntoa(orig_dst_addr.sin_addr));
	
	write_log(LOG_FILE, "[WARNING] Possible Port Scan from [%s:%d -> proxy -> %s:%d] "
		  "passed as LOOSE_PORT_SCAN is enabled",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
		  tmp_str, ntohs(orig_dst_addr.sin_addr), OSCAR);
	
	free(tmp_str);
    }
#else
    if (ntohs(orig_dst_addr.sin_port) != OSCAR) {
	
	tmp_str = strdup(inet_ntoa(orig_dst_addr.sin_addr));
	
	write_log(LOG_FILE, "[WARNING] Possible Port Scan from [%s:%d -> proxy -> %s:%d] "
		  "ignored as destination port is not %d",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
		  tmp_str, ntohs(orig_dst_addr.sin_port), OSCAR);
	
	free(tmp_str);
	close(new_client);
	return;
    }
#endif
    
    // Open the onward portion of the connection (and log errors)
    new_server = net_connect_to(ntohs(orig_dst_addr.sin_port), inet_ntoa(orig_dst_addr.sin_addr));
    if (new_server == -1) {
	
	tmp_str = strdup(inet_ntoa(orig_dst_addr.sin_addr));
	
	write_log(LOG_FILE, "Connection [%s:%d -> proxy -> %s:%d] failed "
		  "[Could not connect to %s:%d]",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
		  tmp_str, ntohs(orig_dst_addr.sin_port),
		  tmp_str, ntohs(orig_dst_addr.sin_port));
	
	free(tmp_str);
	close(new_client);
	return;
    }
    
    // Setup the tracking structure
    ts = malloc(sizeof(user_track_struct));
    memset(ts, 0, sizeof(user_track_struct));
    ts->magic = ts_magic ++;
    ts->next = user_list.next;
    user_list.next = ts;
    
    pt = pts_allocate();
    if (!pt) {
	
	tmp_str = strdup(inet_ntoa(orig_dst_addr.sin_addr));
	
	write_log(LOG_FILE, "Connection [%s:%d -> proxy -> %s:%d] failed "
		  "[Could not allocate a tracking structure]",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
		  tmp_str, ntohs(orig_dst_addr.sin_port));
	free(tmp_str);
	
	ts->cleanup = 1;
	close(new_client);
	close(new_server);
	return;
    }
    
    pt->sockets[0] = new_client;
    pt->sockets[1] = new_server;
    pt->filter[0] = &oscar_tx;
    pt->filter[1] = &oscar_rx;
    pt->user  = 1;
    pt->magic = ts->magic;
    
    pt->next = pt_list.next;
    pt_list.next = pt;
    
    // Can't do it in one go because of the static buffer used in inet_ntoa!
    tmp_str = strdup(inet_ntoa(orig_dst_addr.sin_addr));
    
    write_log(LOG_FILE, "[CONN] OSCAR proxy extablished [%s:%d -> %s:%d] "
	      "[magic=%d]",
	      inet_ntoa(src_addr.sin_addr),
	      ntohs(src_addr.sin_port),
	      tmp_str,
	      ntohs(orig_dst_addr.sin_port),
	      pt->magic);
    free(tmp_str);
}

/*-------------------------------------------------------------------------
  void net_open_aim_dcc(listen_track_struct *lt)
  
  Open the direct connection attempt from the remote party, and allocate
  the dcc structure.  This routine might -not- open the second socket back
  to the inside of the LAN, as without knowing more about the incoming
  request, we have no idea which of the internal hosts we should be
  connecting to.

  However, adhering to the 'nicety' of waiting for the remote end to
  tell us something about itself can lead to the situation where the
  remote end looks connected, but the local end is still waiting.  To
  avoid this, at least on fairly quiet proxies, if there is only 1 DCC
  pending, and LAX_SECURITY is set, we make the onward connection here.
-------------------------------------------------------------------------*/
void net_open_aim_dcc(listen_track_struct *lt)
{
    int new_remote;
    proxy_track_struct *pt;
    struct sockaddr_in src_addr;
    int salen;
    
    // Open the pending connection....
    salen = sizeof(src_addr);
    new_remote = net_open_connect(lt->socket, &src_addr, &salen);
    if (new_remote == -1)
	return;
    
    // Setup the tracking structure
    pt = pts_allocate();
    if (!pt) {
	write_log(LOG_FILE, "Couldnt allocate track struct [%s:%d -> proxy:%d] closed",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port), DCC);
	close(new_remote);
	return;
    }
    
    pt->sockets[TS_SERVER] = new_remote;
    pt->filter[0] = &dcc_tx;
    pt->filter[1] = &dcc_rx;
    
    pt->next = pt_list.next;
    pt_list.next = pt;
    
#if LAX_SECURITY
    
    {
	pending_dc_struct *pdc, *first_pdc;
	int count = 0, total_count = 0;
	
	pdc = pending_dc_list;
	first_pdc = NULL;
	
	while(pdc) {
	    if (pdc->cleanup == 0) {
		
		total_count ++;
		
		if (pdc->port == DCC) {
		    count ++;
		    first_pdc = pdc;
		}
	    }
	    pdc = pdc->next;
	}
	
	// Blow off the connection if there's no pending list...
	if (count == 0) {
	    if (total_count == 0) {
		
		write_log(LOG_FILE, "[WARNING] Possible scan as no pending connections [%s:%d -> proxy:%d]",
			  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
			  DCC);
	    } else {
		write_log(LOG_FILE, "[WARNING] Possible scan or bug as no pending DCCs, but have %d pending connections [%s:%d -> proxy:%d]",
			  total_count,
			  inet_ntoa(src_addr.sin_addr),
			  ntohs(src_addr.sin_port),
			  DCC);
	    }
	    pt->cleanup = 1;
	    close(new_remote);
	    return;
	}
	
	if (count == 1) {
	    
	    proxy_track_struct *pts;
	    
	    pts = pt_list.next;
	    while(pts) {
		
		if ((pts->magic == first_pdc->magic) &&
		    (pts->user == 1)) {
		    
		    int client;
		    struct sockaddr_in cli_addr;
		    int salen;
		    char *tmp_str;
		    
		    // Read the source IP for the client connection.
		    salen = sizeof(struct sockaddr_in);
		    if (getpeername(pts->sockets[TS_CLIENT],
				    (struct sockaddr*) &cli_addr,
				    &salen) == -1) {
			goto next_pts;
		    }
		    
		    client = net_connect_to(DCC, inet_ntoa(cli_addr.sin_addr));
		    
		    pt->sockets[TS_CLIENT] = client;
		    
		    // Can't do it in one go because of the static buffer used in inet_ntoa!
		    tmp_str = strdup(inet_ntoa(cli_addr.sin_addr));
		    
		    write_log(LOG_FILE, "Loose Proxy connection established [%s:%d -> proxy -> %s:%d]",
			      inet_ntoa(src_addr.sin_addr),
			      ntohs(src_addr.sin_port),
			      tmp_str,
			      DCC);
		    free(tmp_str);
		    
		    first_pdc->cleanup = 1;
		    return;
		}
	    next_pts:
		pts = pts->next;
	    }
	    
	    write_log(LOG_FILE, "[FATAL] State Machine Problem! Should have found a DCC match!");
	    
	    // Clean up the DS.
	    pt->cleanup = 1;
	    return;
	}
	
	write_log(LOG_FILE, "DCC connection half-openned [%s:%d -> proxy:%d] Waiting for data [%d pending]",
		  inet_ntoa(src_addr.sin_addr),
		  ntohs(src_addr.sin_port),
		  DCC, count);
	
	return;
    }
    
#endif
    
    write_log(LOG_FILE, "DCC connection half-openned [%s:%d -> proxy:%d] Waiting for data",
	      inet_ntoa(src_addr.sin_addr),
	      ntohs(src_addr.sin_port),
	      DCC);
}

/*-------------------------------------------------------------------------
  void net_open_aim_share(listen_track_struct *lt)
  
  Open the direct share connection attempt from the remote party, and alloc
  the generic proxy_track struct.  This routine does -not- open the second
  socket back to the inside of the LAN, unless LAX_SECURITY has been used.

  Unless we're willing to assume that if we have only 1 connection pending,
  this connection is the right one, then we need to wait for the remote
  party to send something on this socket.
-------------------------------------------------------------------------*/
void net_open_aim_share(listen_track_struct *lt)
{
    int new_remote;
    proxy_track_struct *pt;
    struct sockaddr_in src_addr;
    int salen;
    
    // Open the pending connection....
    salen = sizeof(src_addr);
    new_remote = net_open_connect(lt->socket, &src_addr, &salen);
    if (new_remote == -1)
	return;
    
    // Setup the tracking structure
    pt = pts_allocate();
    if (!pt) {
	write_log(LOG_FILE, "Couldnt allocate track struct [%s:%d -> proxy:%d] closed",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port), SHARE);
	close(new_remote);
	return;
    }
    
    pt->sockets[1] = new_remote;

    pt->filter[0] = &share_tx;
    pt->filter[1] = &share_rx;
    
    // Add the proxy connection to the list of sockets.
    pt->next = pt_list.next;
    pt_list.next = pt;

#if LAX_SECURITY

    {
	pending_dc_struct *pdc, *first_pdc;
	int count = 0, total_count = 0;
	
	pdc = pending_dc_list;
	first_pdc = NULL;
	
	while(pdc) {
	    if (pdc->cleanup == 0) {
		
		total_count ++;
		
		if (pdc->port == SHARE) {
		    count ++;
		    first_pdc = pdc;
		}
	    }
	    pdc = pdc->next;
	}
	
	// Blow off the connection if there's no pending list...
	if (count == 0) {
	    if (total_count == 0) {
		
		write_log(LOG_FILE, "[WARNING] Possible scan as no pending connections [%s:%d -> proxy:%d]",
			  inet_ntoa(src_addr.sin_addr),
			  ntohs(src_addr.sin_port),
			  SHARE);
	    } else {
		write_log(LOG_FILE, "[WARNING] Possible scan or bug as no pending SHAREs, but have %d pending connections [%s:%d -> proxy:%d]",
			  total_count,
			  inet_ntoa(src_addr.sin_addr),
			  ntohs(src_addr.sin_port),
			  SHARE);
	    }
	    pt->cleanup = 1;
	    close(new_remote);
	    return;
	}
	
	if (count == 1) {
	    
	    proxy_track_struct *pts;
	    
	    pts = pt_list.next;
	    while(pts) {
		
		if ((pts->magic == first_pdc->magic) &&
		    (pts->user == 1)) {
		    
		    int client;
		    struct sockaddr_in cli_addr;
		    int salen;
		    char *tmp_str;
		    
		    // Read the source IP for the client connection.
		    salen = sizeof(struct sockaddr_in);
		    if (getpeername(pts->sockets[TS_CLIENT], (struct sockaddr*) &cli_addr, &salen) == -1) {
			goto next_pts;
		    }
		    
		    client = net_connect_to(SHARE, inet_ntoa(cli_addr.sin_addr));
		    
		    pt->sockets[0] = client;
		    
		    // Can't do it in one go because of the static buffer used in inet_ntoa!
		    tmp_str = strdup(inet_ntoa(cli_addr.sin_addr));
		    
		    write_log(LOG_FILE, "Loose Proxy connection established [%s:%d -> proxy -> %s:%d]",
			      inet_ntoa(src_addr.sin_addr),
			      ntohs(src_addr.sin_port),
			      tmp_str,
			      SHARE);
		    free(tmp_str);

		    first_pdc->cleanup = 1;
		    return;
		}
	    next_pts:
		pts = pts->next;
	    }
	    
	    write_log(LOG_FILE, "[FATAL] State Machine Problem! Should have found a SHARE match!");
	    
	    // Clean up the connection as something fishy just happenned.
	    pt->cleanup = 1;
	    return;
	}
	
	write_log(LOG_FILE, "Share connection half-openned [%s:%d -> proxy:%d] Waiting for data [%d pending]",
		  inet_ntoa(src_addr.sin_addr),
		  ntohs(src_addr.sin_port),
		  SHARE, count);
	
	return;
    }
    
#endif
    
    write_log(LOG_FILE, "Share connection half-openned [%s:%d -> proxy:%d] Waiting for data",
	      inet_ntoa(src_addr.sin_addr),
	      ntohs(src_addr.sin_port),
	      SHARE);
}

/*-------------------------------------------------------------------------
  void* net_open_msn(listen_track_struct *lt)
  
  Open the connection attempt from the LAN client, create the onward
  socket for the proxying, and allocate a tracking structure.
  
  If any steps fail, the client connection is terminated.
-------------------------------------------------------------------------*/
// Open a new client
void net_open_msn(listen_track_struct *lt)
{
    int new_client, new_server;
    user_track_struct  *ts;
    proxy_track_struct *pt;
    struct sockaddr_in orig_dst_addr, src_addr;
    int salen;
    char *tmp_str;
    
    // Open the pending connection....
    salen = sizeof(src_addr);
    new_client = net_open_connect(lt->socket, &src_addr, &salen);
    if (new_client == -1)
	return;
    
    // Read the original destination....
    salen = sizeof(orig_dst_addr);
    if (gettranssockname(new_client, (struct sockaddr*) &orig_dst_addr, &salen) == -1) {
	close(new_client);
	return;
    }

    if (ntohs(orig_dst_addr.sin_port) != lt->port) {
	write_log(LOG_FILE, "[WARNING] Possible Port Scan from [%s:%d] ignored",
	       inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port));
	close(new_client);
	return;
    }
    
    // Open the onward portion of the connection....
    new_server = net_connect_to(ntohs(orig_dst_addr.sin_port),
				inet_ntoa(orig_dst_addr.sin_addr));
    if (new_server == -1) {

	tmp_str = strdup(inet_ntoa(orig_dst_addr.sin_addr));
	
	write_log(LOG_FILE, "Onward Connect [%s:%d -> proxy -> %s:%d] failed",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
		  tmp_str, ntohs(orig_dst_addr.sin_port));
	
	free(tmp_str);
	close(new_client);
	return;
    }
    
    // Setup the tracking structure
    ts = malloc(sizeof(user_track_struct));
    memset(ts, 0, sizeof(user_track_struct));
    ts->magic = ts_magic ++;
    ts->next = user_list.next;
    user_list.next = ts;
    
    pt = pts_allocate();
    if (!pt) {
	tmp_str = strdup(inet_ntoa(orig_dst_addr.sin_addr));
	
	write_log(LOG_FILE, "Couldnt allocate track struct [%s:%d -> proxy -> %s:%d] closed",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
		  tmp_str, ntohs(orig_dst_addr.sin_port));
	free(tmp_str);
	
	ts->cleanup = 1;
	close(new_client);
	close(new_server);
	return;
    }
    
    pt->sockets[0] = new_client;
    pt->sockets[1] = new_server;
    pt->filter[0] = &msn_tx;
    pt->filter[1] = &msn_rx;
    pt->user  = 1;
    pt->magic = ts->magic;
    
    pt->next = pt_list.next;
    pt_list.next = pt;
    
    // Can't do it in one go because of the static buffer used in inet_ntoa!
    tmp_str = strdup(inet_ntoa(orig_dst_addr.sin_addr));
    
    write_log(LOG_FILE, "MSN Proxy connection established [%s:%d -> %s:%d] [magic=%d]",
	      inet_ntoa(src_addr.sin_addr),
	      ntohs(src_addr.sin_port),
	      tmp_str,
	      ntohs(orig_dst_addr.sin_port),
	      pt->magic);
    free(tmp_str);
}

/*-------------------------------------------------------------------------
  void net_open_msnrx(listen_track_struct *lt)
  
  Open the direct connection attempt from the remote party, and allocate
  the dcc structure.  This routine might -not- open the second socket back
  to the inside of the LAN, as without knowing more about the incoming
  request, we have no idea which of the internal hosts we should be
  connecting to.

  TODO: Complete this!!!!
-------------------------------------------------------------------------*/
void net_open_msnrx(listen_track_struct *lt)
{
    int new_remote;
    proxy_track_struct *pt;
    struct sockaddr_in src_addr;
    pending_msn_struct *pms, *first;
    int salen, count;
    
    // Open the pending connection....
    salen = sizeof(src_addr);
    new_remote = net_open_connect(lt->socket, &src_addr, &salen);
    if (new_remote == -1)
	return;
    
    // Setup the tracking structure
    pt = pts_allocate();
    if (!pt) {
	close(new_remote);
	return;
    }

    pt->sockets[TS_SERVER] = new_remote;
    pt->filter[0] = &msnrx_tx;
    pt->filter[1] = &msnrx_rx;
    
    pt->next = pt_list.next;
    pt_list.next = pt;
    
    pms = pending_msn_list;
    first = NULL;
    count = 0;

    while(pms) {

	if (pms->cleanup == 0) {
	    count ++;
	    first = pms;
	}

	pms = pms->next;
    }

    if (count == 1) {

	// Lax openning

	char ipaddr[32];

	sprintf(ipaddr, "%d.%d.%d.%d",
		first->ip[0], first->ip[1], first->ip[2], first->ip[3]);
	
	pt->sockets[TS_CLIENT] = net_connect_to(first->client_port, ipaddr);

	write_log(LOG_FILE, "Lax MSN Rx Accepted %s:%d -> proxy:%d -> %s:%d",
		  inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port),
		  lt->port,
		  ipaddr, first->client_port);
	
	first->cleanup = 1;
	
	return;
    }

    write_log(LOG_FILE, "New MSN Rx Accepted %s:%d -> proxy:%d [Waiting For Data] [%d pending]",
	      inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port), lt->port, count);
}

/*-------------------------------------------------------------------------
  unsigned int fp_process(unsigned char *buff,
                          unsigned int len,
			  unsigned int max,
			  filter_chain *fc,
			  proxy_track_struct *pt)

  Pass the packet through the filter chain.  Nice and simple really.

  A NULL filter chain is valid, and mean 'pass unchanged'
  
-------------------------------------------------------------------------*/

unsigned int fp_process(unsigned char *buff,
                        unsigned int len, unsigned int max,
			filter_chain *fc,
			proxy_track_struct *pt)
{
    while(fc && fc->proc) {

	len = fc->proc(buff, len, max, pt);

	fc = fc->next;

    }

    return len;
}

unsigned int fp_oscar_tx_file_setup(unsigned char *buff,
			 unsigned int len, unsigned int max,
			 proxy_track_struct *pt)
{
    // Not long enough for us to check properly...
    if (len < 26)
	return len;
    
    // Check its long enough for the TLV_TYPE
    if (len < AIM_BUD_NEXT(buff) + 2)
	return len;
    
    if (len < 107)
	return len;

    // Does it look like a potential Direct Connection setup packet ?
    if ((AIM_CMDSTART(buff)        == 0x2a) &&
	(AIM_CHANNEL(buff)         == 0x02) &&
	(AIM_FNAC_FAMILY(buff)     == 0x0004) &&
	(AIM_FNAC_SUBTYPE(buff)    == 0x0006) &&
	(AIM_FNAC_FLAGS(buff)      == 0x0000) &&
	(AIM_SNAC_CHAN(buff)       == 0x0002) &&
	(AIM_SNAC_TLV_TYPE(buff)   == 0x0005)) {

	int ip[4], ip_offset, src_ip[4];
	struct sockaddr_in src_addr;
	int salen;
	unsigned char scan[14];
	pending_dc_struct *pdc;

	// Read the source IP for the client connection.
	// We use this for extra sanity checks in locating the client IP
	// in the setup packet
	salen = sizeof(struct sockaddr_in);
	if (getpeername(pt->sockets[TS_CLIENT], (struct sockaddr*) &src_addr, &salen) == -1) {
	    return len;
	}
	
	sscanf(inet_ntoa(src_addr.sin_addr), "%d.%d.%d.%d", &src_ip[0], &src_ip[1], &src_ip[2], &src_ip[3]);

	// Look for a sequence in the packet that is 00 03 00 04 IP IP IP IP 00 05 00 02 15 BE
	scan[0] = 0x00;
	scan[1] = 0x03;
	scan[2] = 0x00;
	scan[3] = 0x04;
	scan[4] = src_ip[0];
	scan[5] = src_ip[1];
	scan[6] = src_ip[2];
	scan[7] = src_ip[3];
	scan[8] = 0x00;
	scan[9] = 0x05;
	scan[10] = 0x00;
	scan[11] = 0x02;
	scan[12] = 0x15; // Port 5566 is 0x15be
	scan[13] = 0xbe;
	
	for (ip_offset = AIM_SNAC_DATA_OFF(buff); ip_offset < len - 16; ip_offset ++)
	    if (memcmp(scan, &buff[ip_offset], 14) == 0)
		break;
	
	// Didn't find the address to overwrite.
	if (ip_offset == len - 16) {

	    // Search for just the IP address leadin and trailer....
	    for (ip_offset = AIM_SNAC_DATA_OFF(buff); ip_offset < len - 16; ip_offset ++)
		if ((memcmp(scan, &buff[ip_offset+0], 4) == 0) &&
		    (memcmp(scan, &buff[ip_offset+8], 2) == 0))
		    break;
	    
	    // This packet didn't appear to be a direct connect setup
	    // so leave it alone
	    if (ip_offset == len - 16) {
		return len;
	    }
	    
#if LOOSE_CLIENT_IP
	    // If strict checking is on, just note the discrepency...
	    write_log(LOG_FILE, "Loose DCC F Source. Expecting %d.%d.%d.%d, "
		      "found %d.%d.%d.%d. Continuing mangle.", 
		      src_ip[0], src_ip[1], src_ip[2], src_ip[3],
		      buff[ip_offset+4], buff[ip_offset+5],
		      buff[ip_offset+6], buff[ip_offset+7]);
#else
	    // If strict checking is on, complain and don't mangle.
	    write_log(LOG_FILE, "Loose DCC F Source. Expecting %d.%d.%d.%d, "
		      "found %d.%d.%d.%d. Not overwriting.", 
		      src_ip[0], src_ip[1], src_ip[2], src_ip[3],
		      buff[ip_offset+4], buff[ip_offset+5],
		      buff[ip_offset+6], buff[ip_offset+7]);
	    return len;
#endif
	}
	
	// Point to the actual address by adjusting over the search preamble.
	ip_offset += 4;
	
	// Okay.  Determine which address to overwrite the client IP with...
	if (outside_ip == NULL) {
	    
	    // Make a best guess here because we weren't told....
	    struct sockaddr_in svr_addr;
	    
	    // Read the source address of the connection to the server
	    // and give up if we can't...
	    salen = sizeof(struct sockaddr_in);
	    if (getsockname(pt->sockets[TS_SERVER], (struct sockaddr*) &svr_addr, &salen) == -1) {
		write_log(LOG_FILE, "Found a DCC F setup, but couldn't get the external IP - dodo mode.");
		return len;
	    }
	    
	    sscanf(inet_ntoa(svr_addr.sin_addr), "%d.%d.%d.%d",
		   &ip[0], &ip[1], &ip[2], &ip[3]);
	    
	} else {
	    
	    if (sscanf(outside_ip, "%d.%d.%d.%d",
		       &ip[0], &ip[1], &ip[2], &ip[3]) != 4) {
		write_log(LOG_FILE, "Bad external proxy IP address specified - [%s] Dodo mode",
			  outside_ip);
		return len;
	    }
	}
	
	// Mangle the data so it has the outside IP address in the right place
	
	buff[ip_offset + 0] = ip[0];
	buff[ip_offset + 1] = ip[1];
	buff[ip_offset + 2] = ip[2];
	buff[ip_offset + 3] = ip[3];
	
	write_log(LOG_FILE,"[MASQ] DCC F Setup IP Changed "
		  "[%d.%d.%d.%d -> %d.%d.%d.%d] [offset=%d] [buddy=%.*s]",
		  src_ip[0], src_ip[1], src_ip[2], src_ip[3],
		  ip[0], ip[1], ip[2], ip[3],
		  ip_offset,
		  AIM_BUD_NAME_LEN(buff), AIM_BUD_NAME(buff));
	
	// Allocate a pending_dc_struct here.
	pdc = malloc(sizeof(pending_dc_struct));
	memset(pdc, 0, sizeof(pending_dc_struct));

	pdc->magic = pt->magic;
	pdc->id[0] = AIM_SNAC_COOKIE(buff)[0];
	pdc->id[1] = AIM_SNAC_COOKIE(buff)[1];
	pdc->id[2] = AIM_SNAC_COOKIE(buff)[2];
	pdc->id[3] = AIM_SNAC_COOKIE(buff)[3];
	pdc->id[4] = AIM_SNAC_COOKIE(buff)[4];
	pdc->id[5] = AIM_SNAC_COOKIE(buff)[5];
	pdc->id[6] = AIM_SNAC_COOKIE(buff)[6];
	pdc->id[7] = AIM_SNAC_COOKIE(buff)[7];

	pdc->port = SHARE;

	pdc->next = pending_dc_list;
	pending_dc_list = pdc;
    }
    
    return len;
}

unsigned int fp_oscar_tx_dcc_setup(unsigned char *buff,
			 unsigned int len, unsigned int max,
			 proxy_track_struct *pt)
{
    // Not long enough for us to check properly...
    if (len < 26)
	return len;
    
    // Check its long enough for the TLV_TYPE
    if (len < AIM_BUD_NEXT(buff) + 2)
	return len;
    
    // Does it look like a Direct Connection setup packet ?
    if ((AIM_CMDSTART(buff)        == 0x2a) &&
	(AIM_CHANNEL(buff)         == 0x02) &&
	(AIM_FNAC_FAMILY(buff)     == 0x0004) &&
	(AIM_FNAC_SUBTYPE(buff)    == 0x0006) &&
	(AIM_FNAC_FLAGS(buff)      == 0x0000) &&
	(AIM_SNAC_CHAN(buff)       == 0x0002) &&
	(AIM_SNAC_TLV_TYPE(buff)   == 0x0005) &&
	(AIM_SNAC_TLV_NEXT(buff)   == 0x0032)) { // 0x32 is the length of a DCC setup...
	
	int ip[4], ip_offset, src_ip[4];
	struct sockaddr_in src_addr;
	int salen, port;
	unsigned char scan[12];
	pending_dc_struct *pdc;

	// Read the source IP for the client connection.
	// We use this for extra sanity checks in locating the client IP in the setup packet
	salen = sizeof(struct sockaddr_in);
	if (getpeername(pt->sockets[TS_CLIENT], (struct sockaddr*) &src_addr, &salen) == -1) {
	    return len;
	}
	
	// Use this to avoid any problems with byte order if we prod
	// directly.
	sscanf(inet_ntoa(src_addr.sin_addr),
	       "%d.%d.%d.%d",
	       &src_ip[0], &src_ip[1], &src_ip[2], &src_ip[3]);

	// Look for a sequence in the packet that is
	// 00 03 00 04 IP IP IP IP 00 05 00 02 PP PP
	scan[0] = 0x00;
	scan[1] = 0x03;
	scan[2] = 0x00;
	scan[3] = 0x04;
	scan[4] = src_ip[0];
	scan[5] = src_ip[1];
	scan[6] = src_ip[2];
	scan[7] = src_ip[3];
	scan[8] = 0x00;
	scan[9] = 0x05;
	scan[10] = 0x00; // Port prefix
	scan[11] = 0x02;
	
	for (ip_offset = AIM_SNAC_DATA_OFF(buff);
	     ip_offset < len - 14;
	     ip_offset ++)
	    if (memcmp(scan, &buff[ip_offset], 12) == 0)
		break;
	
	// Didn't find the address to overwrite.
	if (ip_offset == len - 14) {
	    
	    // Search for just the IP address leadin and trailer....
	    for (ip_offset = AIM_SNAC_DATA_OFF(buff);
		 ip_offset < len - 14;
		 ip_offset ++)
		if ((memcmp(scan, &buff[ip_offset+0], 4) == 0) &&
		    (memcmp(scan, &buff[ip_offset+8], 2) == 0))
		    break;
	    
	    // Complain if we still didn't find it...
	    if (ip_offset == len - 14) {
		
		write_log(LOG_FILE, "Possible DCC, unable to locate a client IP record! (even loosely) :");
		dump_packet(buff, len);
		
		return len;
	    }
	    
#if LOOSE_CLIENT_IP
	    // If strict checking is on, just note the discrepency...
	    write_log(LOG_FILE, "Loose DCC Source. Expecting %d.%d.%d.%d, "
		      "found %d.%d.%d.%d. Continuing mangle.", 
		      src_ip[0], src_ip[1], src_ip[2], src_ip[3],
		      buff[ip_offset+4], buff[ip_offset+5],
		      buff[ip_offset+6], buff[ip_offset+7]);
#else
	    // If strict checking is on, complain and don't mangle.
	    write_log(LOG_FILE, "Loose DCC Source. Expecting %d.%d.%d.%d, "
		      "found %d.%d.%d.%d. Not overwriting.", 
		      src_ip[0], src_ip[1], src_ip[2], src_ip[3],
		      buff[ip_offset+4], buff[ip_offset+5],
		      buff[ip_offset+6], buff[ip_offset+7]);
	    return len;
#endif
	}
	
	// Point to the actual address by adjusting over the search preamble.
	ip_offset += 4;

	// Determine which address to overwrite the client IP with...
	if (outside_ip == NULL) {
	    
	    // Make a best guess here because we weren't told....
	    struct sockaddr_in svr_addr;
	    
	    // Read the source address of the connection to the server....
	    // Give up if we can't...
	    salen = sizeof(struct sockaddr_in);
	    if (getsockname(pt->sockets[TS_SERVER], (struct sockaddr*) &svr_addr, &salen) == -1) {
		write_log(LOG_FILE, "Found a DC setup, but couldn't work out the external IP.");
		return len;
	    }
	    
	    sscanf(inet_ntoa(svr_addr.sin_addr), "%d.%d.%d.%d",
		   &ip[0], &ip[1], &ip[2], &ip[3]);
	    
	} else {
	    if (sscanf(outside_ip, "%d.%d.%d.%d",
		       &ip[0], &ip[1], &ip[2], &ip[3]) != 4) {
		write_log(LOG_FILE, "Bad external proxy IP address specified - [%s] Dodo non-mangling mode.",
			  outside_ip);
		return len;
	    }
	}
	
	// Mangle the bufferred data so it has the outside IP address in the right place
	
	buff[ip_offset + 0] = ip[0];
	buff[ip_offset + 1] = ip[1];
	buff[ip_offset + 2] = ip[2];
	buff[ip_offset + 3] = ip[3];
	
	port = ((buff[ip_offset + 8] << 8) + buff[ip_offset + 9]);
	
	// Allocate a pending_dc_struct here.
	pdc = malloc(sizeof(pending_dc_struct));
	memset(pdc, 0, sizeof(pending_dc_struct));
	
	pdc->magic = pt->magic;
	pdc->id[0] = AIM_SNAC_COOKIE(buff)[0];
	pdc->id[1] = AIM_SNAC_COOKIE(buff)[1];
	pdc->id[2] = AIM_SNAC_COOKIE(buff)[2];
	pdc->id[3] = AIM_SNAC_COOKIE(buff)[3];
	pdc->id[4] = AIM_SNAC_COOKIE(buff)[4];
	pdc->id[5] = AIM_SNAC_COOKIE(buff)[5];
	pdc->id[6] = AIM_SNAC_COOKIE(buff)[6];
	pdc->id[7] = AIM_SNAC_COOKIE(buff)[7];

	// The DCC is setup using the OSCAR port.
	if (port == OSCAR)
	    pdc->port = DCC;
	else
	    pdc->port = port;
	
	pdc->next = pending_dc_list;
	pending_dc_list = pdc;

	write_log(LOG_FILE, "DCC Setup IP Changed...%d.%d.%d.%d:%d -> %d.%d.%d.%d:%d "
		  "[offset=%d] [buddy=%.*s] [listen port=%d]",
		  src_ip[0], src_ip[1], src_ip[2], src_ip[3], port,
		  ip[0], ip[1], ip[2], ip[3], port,
		  ip_offset, AIM_BUD_NAME_LEN(buff), AIM_BUD_NAME(buff),
		  pdc->port);
    }
    
    return len;
}

unsigned int fp_oscar_rx(unsigned char *buff,
			 unsigned int len, unsigned int max,
			 proxy_track_struct *pt)
{
    // Not long enough for us to check properly...
    if (len < 26)
	return len;
    
    // Check its long enough for the TLV_TYPE
    if (len < AIM_BUD_NEXT(buff) + 2)
	return len;
    
    // Does it look like a Direct Connection Share setup packet ?
    if ((AIM_CMDSTART(buff)        == 0x2a) &&
	(AIM_CHANNEL(buff)         == 0x02) &&
	(AIM_FNAC_FAMILY(buff)     == 0x0004) &&
	(AIM_FNAC_SUBTYPE(buff)    == 0x0007) &&
	(AIM_FNAC_FLAGS(buff)      == 0x0000) &&
	(AIM_SNAC_CHAN(buff)       == 0x0002)) {

	//	printf("Rx DCS ? [buddy=%.*s] Chan=2, Fam=4, ST=7 [%d %d]",
	//	       AIM_BUD_NAME_LEN(buff), AIM_BUD_NAME(buff),
	//	       AIM_SNAC_TLV_TYPE(buff), AIM_SNAC_TLV_NEXT(buff));
    }
    return len;
}

/*-------------------------------------------------------------------------
  unsigned int fp_dcc_rx(unsigned char *buff,
                         unsigned int len, unsigned int max,
		         proxy_track_struct *pt)
  
  This function examines direct connection setup packets arriving back
  from the remote party.  If it looks like a reasonable request, we
  search the pending DC list, and if a match can be found, we open the
  return socket to the client.
  
-------------------------------------------------------------------------*/

unsigned int fp_dcc_rx(unsigned char *buff,
		       unsigned int len, unsigned int max,
		       proxy_track_struct *pt)
{
    pending_dc_struct *pdc;
    struct sockaddr_in rem_addr;
    int salen;
    
    // If the onward connection is established, we're all done
    if (pt->sockets[0] != -1)
	return len;
    
    // Read the source IP for the client connection.
    // We use this for extra sanity checks
    salen = sizeof(struct sockaddr_in);
    if (getpeername(pt->sockets[TS_SERVER], (struct sockaddr*) &rem_addr, &salen) == -1) {
	return 0;
    }
    
    write_log(LOG_FILE, "DirectConnect Data from [%s:%d length=%d]",
	      inet_ntoa(rem_addr.sin_addr), ntohs(rem_addr.sin_port), len);
    
    if (len < 20) {
	return 0;
    }
    
    dump_packet(buff, len);
    
    pdc = pending_dc_list;
    
    while(pdc) {

	if ((pdc->cleanup == 0) &&
	    (pdc->id[0] == DCC_SNAC_COOKIE(buff)[0]) &&
	    (pdc->id[1] == DCC_SNAC_COOKIE(buff)[1]) &&
	    (pdc->id[2] == DCC_SNAC_COOKIE(buff)[2]) &&
	    (pdc->id[3] == DCC_SNAC_COOKIE(buff)[3]) &&
	    (pdc->id[4] == DCC_SNAC_COOKIE(buff)[4]) &&
	    (pdc->id[5] == DCC_SNAC_COOKIE(buff)[5]) &&
	    (pdc->id[6] == DCC_SNAC_COOKIE(buff)[6]) &&
	    (pdc->id[7] == DCC_SNAC_COOKIE(buff)[7])) {

	    proxy_track_struct *pts;

	    pts = pt_list.next;
	    while(pts) {

		if ((pts->magic == pdc->magic) &&
		    (pts->user == 1)) {
		    
		    int client;
		    struct sockaddr_in src_addr;
		    int salen;
		    char *tmp_str;
		    
		    // Read the source IP for the client connection.
		    salen = sizeof(struct sockaddr_in);
		    if (getpeername(pts->sockets[TS_CLIENT],
				    (struct sockaddr*) &src_addr,
				    &salen) == -1) {
			goto next_pts;
		    }
		    
		    client = net_connect_to(DCC, inet_ntoa(src_addr.sin_addr));
		    
		    pt->sockets[TS_CLIENT] = client;
		    
		    tmp_str = strdup(inet_ntoa(src_addr.sin_addr));
		    
		    write_log(LOG_FILE,"DCC Established [%s:%d -> proxy:%d -> %s:%d]",
			      inet_ntoa(rem_addr.sin_addr),
			      ntohs(rem_addr.sin_port), DCC,
			      tmp_str, DCC);
		    
		    free(tmp_str);
		    
		    pdc->cleanup = 1;
		    return len;
		}
	    next_pts:
		pts = pts->next;
	    }
	    
	    write_log(LOG_FILE, "[FATAL] State Machine Problem! Should have found a direct client match!");
	    
	    pt->cleanup = 1;
	    return 0;
	}
	
	pdc = pdc->next;
    };

    return 0;
}

unsigned int fp_msn_tx_dcc_setup(unsigned char *buff,
			 unsigned int len, unsigned int max,
			 proxy_track_struct *pt)
{
    int port, cookie;

    int ip[4], ip_offset, src_ip[4];
    struct sockaddr_in src_addr;
    int salen;
    pending_msn_struct *pdc;
    char copy_msg[2048], *pHdrs[32], *pMsgs[32], *scan, tmpbuf[256];
    int hdr_fields, msg_fields, find;
    
    // Not long enough for a dcc setup request or too long...
    if (len < 85 || len >= sizeof(copy_msg))
	return len;
    
    memcpy(copy_msg, buff, len);
    copy_msg[len] = '\0';

    hdr_fields = 0;
    
    for (scan = copy_msg;
	 (scan != NULL) &&
	 (scan[0] != '\0') &&
	     (scan[0] != '\r') &&
	     (scan[1] != '\n');) {

	pHdrs[hdr_fields] = scan;
	hdr_fields ++;
	if (hdr_fields == 32)
	    return len;
	
	scan = strstr(scan, "\r\n");
	if (scan) {
	    *scan = '\0'; // NULL terminate the pHdr
	    scan += 2;    // Shuffle past the \r\n
	}
    }
    
    // Badly formed buffer
    if (!scan || scan[0] == '\0')
	return len;

    scan += 2; // Skip the \r\n separator

    msg_fields = 0;
    
    for (;
	 (scan != NULL) &&
	 (scan[0] != '\0') &&
	     (scan[0] != '\r') &&
	     (scan[1] != '\n');) {

	pMsgs[msg_fields] = scan;
	msg_fields ++;
	if (msg_fields == 32)
	    return len;
	
	scan = strstr(scan, "\r\n");
	if (scan) {
	    *scan = '\0'; // NULL terminate the pMsg
	    scan += 2;    // Shuffle past the \r\n
	}
    }
    
    // Badly formed buffer
    if (!scan || scan[0] == '\0')
	return len;

    // Look for the invite / accepted message....

    for (find = 0; find < hdr_fields; find ++) {
	if (strcmp(pHdrs[find],
		   "Content-Type: text/x-msmsgsinvite; charset=UTF-8") == 0)
	    break;
    }
    
    if (find == hdr_fields)
	return len;

    for (find = 0; find < msg_fields; find ++) {
	if (strcmp(pMsgs[find],
		   "Invitation-Command: ACCEPT") == 0)
	    break;
    }

    if (find == msg_fields)
	return len;

    // Find the IP address header in the packet
    for (find = 0; find < msg_fields; find ++) {
	if (strncmp(pMsgs[find], "IP-Address: ",
		    strlen("IP-Address: ")) == 0) {

	    if (sscanf(pMsgs[find], "IP-Address: %d.%d.%d.%d",
		       &ip[0], &ip[1], &ip[2], &ip[3]) != 4) {
		write_log(LOG_FILE, "MSN DCC: Badly formed IP Address - dodo mode");
		return len;
	    }
	    
	    break;
	}
    }
    
    if (find == msg_fields)
	return len;
    
    // Find the IP address header in the packet
    for (find = 0; find < msg_fields; find ++) {
	if (strncmp(pMsgs[find], "Port: ",
		    strlen("Port: ")) == 0) {
	    
	    if (sscanf(pMsgs[find], "Port: %d",
		       &port) != 1) {
		write_log(LOG_FILE, "MSN DCC: Badly formed IP port specifier - dodo mode");
		return len;
	    }
	    
	    break;
	}
    }
    
    if (find == msg_fields)
	return len;
    
    // Find the AuthCookie in the packet
    for (find = 0; find < msg_fields; find ++) {
	if (strncmp(pMsgs[find], "AuthCookie: ",
		    strlen("AuthCookie: ")) == 0) {

	    if (sscanf(pMsgs[find], "AuthCookie: %d",
		       &cookie) != 1) {
		write_log(LOG_FILE, "MSN DCC: no cookie found - dodo mode");
		return len;
	    }
	    
	    break;
	}
    }
    
    if (find == msg_fields)
	return len;
    
    // Read the source IP for the client connection.
    // We use this for extra sanity checks in locating the client IP in
    // the setup packet
    salen = sizeof(struct sockaddr_in);
    if (getpeername(pt->sockets[TS_CLIENT], (struct sockaddr*) &src_addr, &salen) == -1) {
	return len;
    }
    
    sscanf(inet_ntoa(src_addr.sin_addr), "%d.%d.%d.%d",
	   &src_ip[0], &src_ip[1], &src_ip[2], &src_ip[3]);
    
    if ((ip[0] != src_ip[0]) || (ip[1] != src_ip[1]) ||
	(ip[2] != src_ip[2]) || (ip[3] != src_ip[3])) {
#if LOOSE_CLIENT_IP
	// If strict checking is on, just note the discrepency...
	write_log(LOG_FILE, "MSN Loose DCC Source. Expecting %d.%d.%d.%d, found %d.%d.%d.%d. Continuing mangle.", 
		  src_ip[0], src_ip[1], src_ip[2], src_ip[3],
		  buff[ip_offset+4], buff[ip_offset+5], buff[ip_offset+6], buff[ip_offset+7]);
#else
	// If strict checking is on, complain and don't mangle.
	write_log(LOG_FILE, "MSN Loose DCC Source. Expecting %d.%d.%d.%d, found %d.%d.%d.%d. Not overwriting.", 
		  src_ip[0], src_ip[1], src_ip[2], src_ip[3],
		  buff[ip_offset+4], buff[ip_offset+5], buff[ip_offset+6], buff[ip_offset+7]);
	return len;
#endif
    }
    
    // Okay.  Now to determine which address to overwrite the client IP with...
    if (outside_ip == NULL) {
	    
	// Make a best guess here because we weren't told....
	struct sockaddr_in svr_addr;
	
	// Read the source address of the server connection...
	salen = sizeof(struct sockaddr_in);
	if (getsockname(pt->sockets[TS_SERVER], (struct sockaddr*) &svr_addr, &salen) == -1) {
	    write_log(LOG_FILE, "MSN Found a DC setup, but couldn't work out the external IP - dodo mode");
	    return len;
	}
	
	sscanf(inet_ntoa(svr_addr.sin_addr), "%d.%d.%d.%d", &ip[0], &ip[1], &ip[2], &ip[3]);
	
    } else {
	sscanf(outside_ip, "%d.%d.%d.%d", &ip[0], &ip[1], &ip[2], &ip[3]);
    }
    
    // Create the new packet, now we have all the data.
    
    write_log(LOG_FILE, "MSN DCC Setup IP Changed...%d.%d.%d.%d:%d -> %d.%d.%d.%d:%d",
	      src_ip[0], src_ip[1], src_ip[2], src_ip[3], port,
	      ip[0], ip[1], ip[2], ip[3], MSNRX);
    
    // copy over the headers
    for (find = 0, scan = buff, len = 0; find < hdr_fields; find++) {
	
	sprintf(scan, "%s\r\n", pHdrs[find]);
	scan = &scan[(strlen(pHdrs[find]) + 2)];
	len += (strlen(pHdrs[find]) + 2);
    }
    
    sprintf(scan, "\r\n");
    scan = &scan[2];
    len += 2;
    
    // Copy over and mangle as required the msg fields
    for (find = 0; find < msg_fields; find ++) {
	
	int done = 0;
	
	if (strncmp(pMsgs[find], "IP-Address: ",
		    strlen("IP-Address: ")) == 0) {
	    
	    sprintf(tmpbuf, "IP-Address: %d.%d.%d.%d",
		    ip[0], ip[1], ip[2], ip[3]);
	    
	    sprintf(scan, "%s\r\n", tmpbuf);
	    scan = &scan[(strlen(tmpbuf) + 2)];
	    len += (strlen(tmpbuf) + 2);
	    done = 1;
	    
	} else if (strncmp(pMsgs[find], "Port: ",
		    strlen("Port: ")) == 0) {
	    
	    sprintf(tmpbuf, "Port: %d", MSNRX);
	    
	    sprintf(scan, "%s\r\n", tmpbuf);
	    scan = &scan[(strlen(tmpbuf) + 2)];
	    len += (strlen(tmpbuf) + 2);
	    done = 1;
	    
	}
	
	if (!done) {
	    // Just copy over the field unchanged
	    sprintf(scan, "%s\r\n", pMsgs[find]);
	    scan = &scan[(strlen(pMsgs[find]) + 2)];
	    len += (strlen(pMsgs[find]) + 2);
	    
	}
    }

    // Finish off the command building
    sprintf(scan, "\r\n");
    scan = &scan[2];
    len += 2;
    
    // Allocate a pending_msn_struct here.
    pdc = malloc(sizeof(pending_msn_struct));
    memset(pdc, 0, sizeof(pending_msn_struct));
    
    pdc->client_port = port;
    pdc->cookie      = cookie;

    pdc->ip[0]       = src_ip[0];
    pdc->ip[1]       = src_ip[1];
    pdc->ip[2]       = src_ip[2];
    pdc->ip[3]       = src_ip[3];
    
    pdc->next = pending_msn_list;
    pending_msn_list = pdc;
    
    return len;
}

/*
 * Read in a filter configuration / control file
 *
 * Also dynamic link any modules when I get that far...
 *
 * For now, just do it statically.
 */
void setup_filters(const char *ini_file)
{
    filter_chain *fc;
    
    // AIM FILTERS

    oscar_tx.proc = fp_oscar_tx_dcc_setup;   // LAN client to AIM filter
    oscar_rx.proc = fp_oscar_rx;   // AIM to LAN client filter

    // Add the file filter to the tx check chain
    fc = malloc(sizeof(filter_chain));
    memset(fc, 0, sizeof(fc));
    fc->proc = fp_oscar_tx_file_setup;
    oscar_tx.next = fc;

    //dcc_tx;     // LAN to Remote filter

    dcc_rx.proc = fp_dcc_rx; // remote to LAN recieve check

    //share_tx;   // LAN to Remote filter
    //share_rx;   // Remote to LAN filter

    // MSN FILTERS

    msn_tx.proc = fp_msn_tx_dcc_setup;  // LAN client to MSG filter
    //msn_rx;

    //msnrx_tx;   // MSN file xfer things
    //msnrx_rx;   
}

/*
 * Read in a filter configuration / control file
 *
 * Also dynamic link any modules when I get that far...
 *
 * For now, just do it statically.
 */
void setup_listenners(const char *ini_file)
{
    listen_track_struct *lt;

    // AIM: OSCAR
    lt = malloc(sizeof(listen_track_struct));
    memset(lt, 0, sizeof(listen_track_struct));
    lt->port   = OSCAR;
    lt->socket = net_listen(lt->port, "0.0.0.0");
    lt->proc   = net_open_aim_oscar;
    lt->next = lt_list.next;
    lt_list.next = lt;
    
    // AIM: INLINE DCC
    lt = malloc(sizeof(listen_track_struct));
    memset(lt, 0, sizeof(listen_track_struct));
    lt->port   = DCC;
    lt->socket = net_listen(lt->port, "0.0.0.0");
    lt->proc   = net_open_aim_dcc;
    lt->next = lt_list.next;
    lt_list.next = lt;
    
    // AIM: FILE SHARING
    lt = malloc(sizeof(listen_track_struct));
    memset(lt, 0, sizeof(listen_track_struct));
    lt->port   = SHARE;
    lt->socket = net_listen(lt->port, "0.0.0.0");
    lt->proc   = net_open_aim_share;
    lt->next = lt_list.next;
    lt_list.next = lt;
    
    // MSN: CLIENT-SERVER
    lt = malloc(sizeof(listen_track_struct));
    memset(lt, 0, sizeof(listen_track_struct));
    lt->port   = MSN;
    lt->socket = net_listen(lt->port, "0.0.0.0");
    lt->proc   = net_open_msn;
    lt->next = lt_list.next;
    lt_list.next = lt;
    
    // MSN: FILE XFER
    lt = malloc(sizeof(listen_track_struct));
    memset(lt, 0, sizeof(listen_track_struct));
    lt->port   = MSNRX;
    lt->socket = net_listen(lt->port, "0.0.0.0");
    lt->proc   = net_open_msnrx;
    lt->next = lt_list.next;
    lt_list.next = lt;
}

#ifdef BUILD_STATIC
int reaim_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
    fd_set rfds, efds, wfds;
    struct timeval tv;
    int sv;
    struct sigaction set_act, old_act;
    proxy_track_struct *pt, *prev_pt;
    listen_track_struct *lt;
    int max_fd, sock;
    
    /* Setup the action table for SIGPIPE - ignore. */
    set_act.sa_handler  = SIG_IGN;
    sigemptyset(&set_act.sa_mask);
    set_act.sa_flags    = 0;
    set_act.sa_restorer = 0;
    sigaction(SIGPIPE, &set_act, &old_act);
    
    if (write_log(LOG_FILE,
		  "[STARTUP] Started AIM-Fix Proxy [%s]",
		  REAIM_VERSION) == 0) {
	fprintf(stderr,
		"[STARTUP] Started AIM-Fix Proxy [%s]\n"
		"[WARNING] Unable to write log to %s.\n"
		"[WARNING] Proxy will continue, but logging will not work.\n",
		REAIM_VERSION, LOG_FILE);
    };
    
    if (argc > 1) {
	outside_ip = argv[1];
	write_log(LOG_FILE, "[STARTUP] Using %s for the external IP address.", outside_ip);
    } else
	write_log(LOG_FILE, "[STARTUP] Auto-detecting the external IP address.");
    
    setup_filters("");
    
    setup_listenners("");
    
    // - Wait for connections to the redirected port
    // - Setup the onward connection to listen
    // - Make onward connection to source
    // - Push around any packets
    
    while(1) {

	// Setup the list of sockets we are monitoring
	FD_ZERO(&rfds);
	FD_ZERO(&efds);
	FD_ZERO(&wfds);
	
	max_fd = 0;

	// Add the listenners to the socket list.
	lt = lt_list.next;
	while(lt != NULL) {
	    
	    if (lt->socket > max_fd)
		max_fd = lt->socket;
	    
	    FD_SET(lt->socket, &rfds);

	    lt = lt->next;
	}

	pt = pt_list.next;
	prev_pt = &pt_list;
	while(pt != NULL) {

	    // Cleanup any proxy connections that are flagged...
	    if (pt->cleanup) {

		proxy_track_struct *clean_pt = pt;
		
		write_log(LOG_FILE, "[Cleanup] proxy_track_struct [magic=%d] [user=%d]",
			  pt->magic, pt->user);

		// Adjust the linked list
		prev_pt->next = pt->next;
		pt = pt->next;

		// Clean up the proxy track structure data
		
		if (clean_pt->data_buff[0])
		    free(clean_pt->data_buff[0]);

		if (clean_pt->data_buff[1])
		    free(clean_pt->data_buff[1]);

		if (clean_pt->sockets[0] != -1)
		    close(clean_pt->sockets[0]);
		if (clean_pt->sockets[1] != -1)
		    close(clean_pt->sockets[1]);

		// If it's the user-associated socket, zap the user
		if (clean_pt->user == 1) {

		    user_track_struct *user_ut = user_list.next;

		    while(user_ut) {
			if (user_ut->magic == clean_pt->magic)
			    user_ut->cleanup = 1;
			
			user_ut = user_ut->next;
		    }
		}
		
		free(clean_pt);
		
		continue;
	    }

	    // Add the sockets to the read/write selection arrays
	    // as appropriate.
	    for (sock = 0; sock < 2; sock ++) {

		if (pt->sockets[sock] != -1) {
		    FD_SET(pt->sockets[sock], &rfds);
		    
		    if (pt->data_len[sock] > 0)
			FD_SET(pt->sockets[sock], &wfds);
		    
		    if (pt->sockets[sock] > max_fd)
			max_fd = pt->sockets[sock];
		}
	    }

	    prev_pt = pt;
	    pt = pt->next;
	}
	
	// Wait for some activity
	tv.tv_usec = 0;
	tv.tv_sec  = 1;
	
	sv = select(max_fd+1,&rfds,&wfds,&efds,&tv);
	switch(sv) {
	    
	case -1:
	    /* Some socket error, maybe */
	    if (errno != EINTR) {
		perror("FATAL: Server socket died: ");
		exit(1);
	    }
	    break;
	case 0:
	    // Just timed out with nothing going on.
	    break;
	    
	default:
	    
	    // Process listenners...
	    lt = lt_list.next;
	    while(lt != NULL) {
		
		if ((lt->socket != -1) &&
		    (FD_ISSET(lt->socket, &rfds)))
		    lt->proc(lt);
		
		lt = lt->next;
	    }
	    
	    // Check generic proxied port pairs
            pt = pt_list.next;
	    while(pt) {

		for (sock = 0; sock < 2; sock ++) {
		    
		    // On Write list ?  We must have pending data for it...
		    if ((pt->sockets[sock] != -1) &&
			(pt->cleanup == 0) &&
			FD_ISSET(pt->sockets[sock], &wfds)) {
			
			int txs;
			
			txs = write(pt->sockets[sock],
				    &pt->data_buff[sock][pt->data_offset[sock]],
				    pt->data_len[sock]);
			if (txs < 0) {
			    if ((errno != EINTR) && (errno != EAGAIN)) {
				write_log(LOG_FILE, "Socket write failure! cleaning up!");
				pt->cleanup = 1;
			    }
			} else {
			    pt->data_len[sock] -= txs;
			    if (pt->data_len[sock] == 0)
				pt->data_offset[sock] = 0;
			    else
				pt->data_offset[sock] += txs;
			}
		    }
		    
		    // On read list, we have data for the other half...
		    //
		    // We only read it if the other half's write buffer is
		    // completely empty.
		    //
		    if ((pt->sockets[sock] != -1) &&
			(pt->cleanup == 0) &&
			FD_ISSET(pt->sockets[sock], &rfds) &&
			(pt->data_len[sock ^ 1] == 0)) {
			
			int rxs;

			// Read into the other half's data buffer.
			rxs = read(pt->sockets[sock],
				   pt->data_buff[sock^1],
				   pt->data_max[sock^1]);
			if (rxs == -1) {
			    if (errno == EAGAIN)
				write_log(LOG_FILE, "FOOIE! FD_ISSET, but read() returned EAGAIN [%d]!",
					  pt->sockets[sock]);
			    if ((errno != EINTR) && (errno != EAGAIN)) {
				pt->cleanup = 1;
			    }
			}
			
			if (rxs == 0) {
			    write_log(LOG_FILE, "socket closed.  cleaning up.....");
			    pt->cleanup = 1;
			}
			
			if (rxs > 0) {
			    
			    pt->data_len[sock^1] =
				fp_process(pt->data_buff[sock^1],
					   rxs,
					   pt->data_max[sock^1],
					   pt->filter[sock],
					   pt);
			}
		    }
		    
		} // End of for(sock...)
		
		pt = pt->next;
		
	    } // End of while(pt)

	    break;
	} // End of switch(sv)
	
    } // End of while(1)
    
    return 0;
}
