#include "flash_config.h"
#include <getopt.h>

// Configuration data, saved in FLASH, used to set/update RedBoot
// normal "configuration" data items.
static struct _config {
    unsigned long len;
    unsigned long key1;
    unsigned char config_data[MAX_CONFIG_DATA-(4*4)];
    unsigned long key2;
    unsigned long cksum;
} *config, *backup_config;
static bool config_ok;
unsigned char fconfig_data[CYGNUM_REDBOOT_FLASH_CONFIG_SIZE];
unsigned char script[CYGNUM_REDBOOT_FLASH_SCRIPT_SIZE];
unsigned long script_timeout;

#define CYGNUM_REDBOOT_FLASH_STRING_SIZE 128
unsigned char esa[128];
unsigned long inventel_wireless_magic = 0x00000000;

#define CONFIG_KEY1    0x0BADFACE
#define CONFIG_KEY2    0xDEADDEAD

#define CONFIG_DONE    0
#define CONFIG_ABORT  -1
#define CONFIG_CHANGED 1
#define CONFIG_OK      2
#define CONFIG_BACK    3
#define CONFIG_BAD     4

// 
// Layout of config data
// Each data item is variable length, with the name, type and dependencies
// encoded into the object.
//  offset   contents
//       0   data type
//       1   length of name (N)
//       2   enable sense
//       3   length of enable key (M)
//       4   key name
//     N+4   enable key
//   M+N+4   data value
//

#define CONFIG_OBJECT_TYPE(dp)          (dp)[0]
#define CONFIG_OBJECT_KEYLEN(dp)        (dp)[1]
#define CONFIG_OBJECT_ENABLE_SENSE(dp)  (dp)[2]
#define CONFIG_OBJECT_ENABLE_KEYLEN(dp) (dp)[3]
#define CONFIG_OBJECT_KEY(dp)           ((dp)+4)
#define CONFIG_OBJECT_ENABLE_KEY(dp)    ((dp)+4+CONFIG_OBJECT_KEYLEN(dp))
#define CONFIG_OBJECT_VALUE(dp)         ((dp)+4+CONFIG_OBJECT_KEYLEN(dp)+CONFIG_OBJECT_ENABLE_KEYLEN(dp))

#define LIST_OPT_LIST_ONLY (1)
#define LIST_OPT_NICKNAMES (2)
#define LIST_OPT_FULLNAMES (4)

static unsigned char *flash_lookup_config(char *key);




#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef CONFIG_LINUX_2_6
#include <mtd/mtd-user.h>
#else
#include <linux/mtd/mtd.h>
#endif


// #define DEBUG

/* for debugging purposes only */
#ifdef DEBUG
#undef DEBUG
#define DEBUG(fmt,args...) { log_printf (LOG_ERROR,"%d: ",__LINE__); log_printf (LOG_ERROR,fmt,## args); }
#else
#undef DEBUG
#define DEBUG(fmt,args...)
#endif


/* error levels */
#define LOG_NORMAL	1
#define LOG_ERROR	2

/* cmd-line flags */
#define FLAG_NONE		0x00
#define FLAG_VERBOSE	0x01
#define FLAG_HELP		0x02
#define FLAG_FILENAME	0x04
#define FLAG_DEVICE		0x08

int flags = FLAG_NONE;
static int dev_fd = -1;

static void log_printf (int level,const char *fmt, ...)
{
   FILE *fp = level == LOG_NORMAL ? stdout : stderr;

   if (flags & FLAG_VERBOSE)
   {
     va_list ap;
     va_start (ap,fmt);
     vfprintf (fp,fmt,ap);
     va_end (ap);
     fflush (fp);
   }
}


void my_exit(int result)
{
  //printf("00:07:3A:FF:FF:FF\n");
    exit(result);
}


static int safe_open (const char *pathname,int flags)
{
   int fd;

   fd = open (pathname,flags);
   if (fd < 0)
	 {
		log_printf (LOG_ERROR,"While trying to open %s",pathname);
		if (flags & O_RDWR)
		  log_printf (LOG_ERROR," for read/write access");
		else if (flags & O_RDONLY)
		  log_printf (LOG_ERROR," for read access");
		else if (flags & O_WRONLY)
		  log_printf (LOG_ERROR," for write access");
		log_printf (LOG_ERROR,": %m\n");
		my_exit (EXIT_FAILURE);
	 }

   return (fd);
}

static void safe_read (int fd,const char *filename,void *buf,size_t count,bool verbose)
{
   ssize_t result;

   result = read (fd,buf,count);
   if (count != result)
	 {
		if (verbose) log_printf (LOG_NORMAL,"\n");
		if (result < 0)
		  {
			 log_printf (LOG_ERROR,"While reading data from %s: %m\n",filename);
			 my_exit (EXIT_FAILURE);
		  }
		log_printf (LOG_ERROR,"Short read count returned while reading from %s\n",filename);
		my_exit (EXIT_FAILURE);
	 }
}


int flash_read(const char *device, unsigned char *dest)
{
   int i,flags = FLAG_NONE;
   size_t result,size,written;
   struct mtd_info_user mtd;
   unsigned char data[0x20000];

   /* get some info about the flash device */
   
   log_printf(LOG_NORMAL,"flash_read device %s for redboot flash configuration\n", device);
   //   dev_fd = safe_open (device,O_SYNC | O_RDWR);
   dev_fd = safe_open (device, O_RDONLY);
   if (ioctl (dev_fd,MEMGETINFO,&mtd) < 0)
	 {
		DEBUG("ioctl(): %m\n");
		log_printf (LOG_ERROR,"This doesn't seem to be a valid MTD flash device!\n");
		my_exit (EXIT_FAILURE);
	 }


   /**********************************
	* verify that flash == file data *
	**********************************/

   //safe_read (dev_fd,device,data,0x20000,flags & FLAG_VERBOSE);
   safe_read (dev_fd,device,dest,0x1000,flags & FLAG_VERBOSE);

   for (i = 0 ; i < 4096 ; i ++)
   {
     //dest[i] = data[0x10000 + i];
     if ((!(i & 0x07)) && (i != 0))
       diag_printf(" ");
     if ((!(i & 0x0F)) && (i != 0))
       diag_printf("\n");
       diag_printf("%02X ", dest[i]);
   }
   diag_printf("\n");
   return 0 ; 
}




/* Return the first occurrence of NEEDLE in HAYSTACK.  */
void *
memmem (haystack, haystack_len, needle, needle_len)
     const void *haystack;
     size_t haystack_len;
     const void *needle;
     size_t needle_len;
{
  const char *begin; 
  const char *const last_possible
    = (const char *) haystack + haystack_len - needle_len;

  if (needle_len == 0)
    /* The first occurrence of the empty string is deemed to occur at
       the beginning of the string.  */
    return (void *) haystack;

  /* Sanity check, otherwise the loop might search through the whole
     memory.  */
  if (haystack_len < needle_len)
    return NULL;

  for (begin = (const char *) haystack; begin <= last_possible; ++begin)
    if (begin[0] == ((const char *) needle)[0] &&
        !memcmp ((const void *) &begin[1],
                 (const void *) ((const char *) needle + 1),
                 needle_len - 1))
      return (void *) begin;

  return NULL;
}









//
// Manage configuration information with the FLASH
//

static int
config_length(int type)
{
    switch (type) {
    case CONFIG_BOOL:
        return sizeof(bool);
    case CONFIG_INT:
        return sizeof(unsigned long);
#ifdef CYGPKG_REDBOOT_NETWORKING
    case CONFIG_IP:
        return sizeof(in_addr_t);
    case CONFIG_ESA:
        // Would like this to be sizeof(enet_addr_t), but that causes much
        // pain since it fouls the alignment of data which follows.
        return 8;
#endif
    case CONFIG_STRING:
        return MAX_STRING_LENGTH;
    case CONFIG_SCRIPT:
        return MAX_SCRIPT_LENGTH;
    default:
        return 0;
    }
}

//
// Retrieve a data object from the data base (in memory copy)
//
bool
flash_get_config(char *key, void *val, int type)
{
    unsigned char *dp;
    void *val_ptr;

    /* inv_xde */
    //diag_printf("flash_get_config : %s\n", key);

    if (!config_ok) return -1;

    if ((dp = flash_lookup_config(key)) != (unsigned char *)NULL) {
        if (CONFIG_OBJECT_TYPE(dp) == type) {
            val_ptr = (void *)CONFIG_OBJECT_VALUE(dp);
            switch (type) {
                // Note: the data may be unaligned in the configuration data
            case CONFIG_BOOL:
                memcpy(val, val_ptr, sizeof(bool));
                break;
            case CONFIG_INT:
                memcpy(val, val_ptr, sizeof(unsigned long));
                break;
#ifdef CYGPKG_REDBOOT_NETWORKING
            case CONFIG_IP:
                memcpy(val, val_ptr, sizeof(in_addr_t));
                break;
            case CONFIG_ESA:
                memcpy(val, val_ptr, sizeof(enet_addr_t));
                break;
#endif
            case CONFIG_STRING:
            case CONFIG_SCRIPT:
                // Just return a pointer to the script/line
                *(unsigned char **)val = (unsigned char *)val_ptr;
                break;
            }
        } else {
            diag_printf("Request for config value '%s' - wrong type\n", key);
        }
        return true;
    }
    return false;
}

//
// Find the configuration entry for a particular key
//
static unsigned char *
flash_lookup_config(char *key)
{
    unsigned char *dp;
    int len;

    if (!config_ok) return (unsigned char *)NULL;

    dp = &config->config_data[0];
    while (dp < &config->config_data[sizeof(config->config_data)]) {
        len = 4 + CONFIG_OBJECT_KEYLEN(dp) + CONFIG_OBJECT_ENABLE_KEYLEN(dp) +
            config_length(CONFIG_OBJECT_TYPE(dp));
        if (strcmp(key, CONFIG_OBJECT_KEY(dp)) == 0) {
            return dp;
        }
        dp += len;
    }
    diag_printf("Can't find config data for '%s'\n", key);
    return false;
}

//
// Attempt to get configuration information from the FLASH.
// If available (i.e. good checksum, etc), initialize "known"
// values for later use.
//
unsigned char *wep_key_dyn = 0;
unsigned char *serial_number_dyn = 0;
unsigned char *h235_key_dyn = 0;


static int
load_flash_config(const char *device)
{
    bool use_boot_script, set_esa, wep_key;
    unsigned char i ;

    config_ok = false;
    //script = (unsigned char *)0;
    //if (!do_flash_init()) return;
    //config = (struct _config *)(workspace_end-sizeof(struct _config));
    config = (struct _config *)fconfig_data;
    //backup_config = (struct _config *)((CYG_ADDRESS)config-sizeof(struct _config));
    //workspace_end = (unsigned char *)backup_config;
#if 0
#define _roundup(n,s) ((((n)+(s-1))/s)*s)
    cfg_size = (block_size > sizeof(struct _config)) ?
        sizeof(struct _config) :
        _roundup(sizeof(struct _config), block_size);
    if (CYGNUM_REDBOOT_FLASH_CONFIG_BLOCK < 0) {
        cfg_base = (void *)((CYG_ADDRESS)flash_end + 1 +
                            (CYGNUM_REDBOOT_FLASH_CONFIG_BLOCK*block_size));
    } else {
        cfg_base = (void *)((CYG_ADDRESS)flash_start + 
                            (CYGNUM_REDBOOT_FLASH_CONFIG_BLOCK*block_size));
    }

/* inv_xde */
#ifdef CYGOPT_REDBOOT_FIS
    if (CYGNUM_REDBOOT_FLASH_CONFIG_BLOCK == CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK)
    {
      cfg_base += block_size / 2;
    
      if (CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK < 0) { // in case fis command not called before
	fis_addr = (void *)((CYG_ADDRESS)flash_end + 1 +
			    (CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK*block_size));
      } else {
	fis_addr = (void *)((CYG_ADDRESS)flash_start + 
			    (CYGNUM_REDBOOT_FIS_DIRECTORY_BLOCK*block_size));
      }
      memcpy(fis_work_block, fis_addr, fisdir_size);
    }
#endif    
#endif
    //memcpy(config, cfg_base, sizeof(struct _config));
    flash_read(device, fconfig_data);
    //printf("cksum : 0x%08X\n", config->cksum);
    if ((crc32((unsigned char *)config, 
               sizeof(struct _config)-sizeof(config->cksum)) != config->cksum) ||
        (config->key1 != CONFIG_KEY1)|| (config->key2 != CONFIG_KEY2)) {
        diag_printf("FLASH configuration checksum error or invalid key\n");
        //config_init();
        return 1;
    }
    config_ok = true;
    flash_get_config("boot_script", &use_boot_script, CONFIG_BOOL);
    if (use_boot_script) {
        flash_get_config("boot_script_data", &script, CONFIG_SCRIPT);
        flash_get_config("boot_script_timeout", &script_timeout, CONFIG_INT);
    }
#ifdef CYGSEM_REDBOOT_VARIABLE_BAUD_RATE
    if (flash_get_config("console_baud_rate", &console_baud_rate, CONFIG_INT)) {
        extern void set_console_baud_rate(int);
        set_console_baud_rate(console_baud_rate);
    }
#endif

    return 0;
}


static void showusage (const char *progname,bool error)
{
  int level = error ? LOG_ERROR : LOG_NORMAL;

  log_printf (level,
	      "\n"
	      "Read Ethernet Mac Address from RedBoot Flash Configuration\n"
	      "Written by Xavier Debreuil <xavier.debreuil@inventel.fr>\n"
	      "\n"
	      "usage: %s [ -v | --verbose ] <device>\n"
	      "       %s -h | --help\n"
	      "\n"
	      "   -h | --help      Show this help message\n"
	      "   -v | --verbose   Show progress reports\n"
	      "   <device>         Flash device to read from (e.g. /dev/mtd0, /dev/mtd1, etc.)\n"
	      "\n",
	      progname,progname);

  exit (error ? EXIT_FAILURE : EXIT_SUCCESS);
}


static char *parameter;
struct option options[] = {
  { "parameter",        1, 0,         'p' },
  { "bdaddr_select",    0, 0,         'b' },
  { "bridee",           0, 0,         'B' },
};

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <getopt.h>
#include <errno.h>

#include "conffile/rg_lock.h"

int main (int argc,char **argv)
{
  const char *progname = "redboot_fconfig_get", *device = NULL;
  int i;	
  int opt, got_parameter = 0;
  bool use_boot_script, set_esa, wep_key, serial_number, h235_key;

  while ((opt = getopt_long_only(argc, argv, "vhpbB", options, NULL)) > 0) {
    switch (opt) {
    case 'v':
      flags |= FLAG_VERBOSE;
      DEBUG("Got FLAG_VERBOSE\n");
      break;
    case 'h':
      flags |= FLAG_HELP | FLAG_VERBOSE;
      DEBUG("Got FLAG_HELP\n");
      break;
    case 'p':
      parameter = optarg;
      //printf("parameter : 0x%08X %s\n", parameter, parameter);
      //printf("parameter : 0x%08X %s\n", optarg, optarg);
      got_parameter = 1;
      break;
    case 'b':
      printf("%d\n", rg_read_bdaddr_selector());
      return 0;
    case 'B':
      printf("%d\n", rg_bridee());
      return 0;
   }
  }
  
  if (optind == argc)
    showusage(progname,true);

  device = argv[optind];

  if (!load_flash_config(device)) {
    if (strcmp(parameter, "eth0_esa") == 0) {
      flash_get_config("eth0_esa", &set_esa, CONFIG_BOOL);
      diag_printf("ETHERNET MAC ADDRESS : ");
      if (set_esa) {
	if (flash_get_config("eth0_esa_data", esa, CONFIG_ESA) == false) {
	  printf("00:07:3A:FF:FF:FF\n");
	}
	else {
	  for (i = 0 ; i < 6 ; i ++) {
	    printf("%02X", esa[i]);
	    if (i != 5)
	      printf(":");
	    else
	      printf("\n");
	  }
	}
      }
      else {
	printf("00:07:3A:FF:FF:FF\n");
      }
    }

    else if (strcmp(parameter, "wep_key") == 0) {
      flash_get_config("wep_key", &wep_key, CONFIG_BOOL);
      diag_printf("WEP KEY : ");
      if (wep_key) {
	if (flash_get_config("wep_key_data", &wep_key_dyn, CONFIG_STRING) == false) {
	  printf("FFFFFFFFFFFFFFFFFFFFFFFFFF\n");
	}
	else {
	  printf("%s\n", wep_key_dyn);
	}
      }
      else {
	printf("FFFFFFFFFFFFFFFFFFFFFFFFFF\n");
      }
    }

    else if (strcmp(parameter, "serial_number") == 0) {
      flash_get_config("serial_number", &serial_number, CONFIG_BOOL);
      diag_printf("SERIAL NUMBER : ");
      if (serial_number) {
	if (flash_get_config("serial_number_data", &serial_number_dyn, CONFIG_STRING) == false) {
	  printf("0000000000000000\n");
	}
	else {
	  printf("%s\n", serial_number_dyn);
	}
      }
      else {
	printf("0000000000000000\n");
      }
    }

    else if (strcmp(parameter, "h235_key") == 0) {
      flash_get_config("h235_key", &h235_key, CONFIG_BOOL);
      diag_printf("H235 KEY : ");
      if (h235_key) {
	if (flash_get_config("h235_key_data", &h235_key_dyn, CONFIG_STRING) == false) {
	  printf("1234567812345678\n");
	}
	else {
	  printf("%s\n", h235_key_dyn);
	}
      }
      else {
	printf("1234567812345678\n");
      }
    }
    
    else if (strcmp(parameter, "voip_prov_password") == 0) {
      flash_get_config("h235_key", &h235_key, CONFIG_BOOL);
      diag_printf("H235 KEY : ");
      if (h235_key) {
	if (flash_get_config("h235_key_data", &h235_key_dyn, CONFIG_STRING) == false) {
	  printf("12345678\n");
	}
	else {
	  char voip_prov_password[9];
	  strncpy(voip_prov_password, h235_key_dyn, 8);
	  voip_prov_password[8] = 0;
	  printf("%s\n", voip_prov_password);
	}
      }
      else {
	printf("1234567812345678\n");
      }
    }


    else if (strcmp(parameter, "inventel_wireless_magic") == 0) {
      inventel_wireless_magic = 0x00000000;
      if (flash_get_config("inventel_wireless_magic", &inventel_wireless_magic, CONFIG_INT) == false) {
	printf("0xFFFFFFFF\n");
	return 1;
      }
      diag_printf("INVENTEL WIRELESS MAGIC : ");
      printf("0x%08X\n", inventel_wireless_magic);
    }


    else if (strcmp(parameter, "target_name_index") == 0) {
      inventel_wireless_magic = 0x00000000;
      if (flash_get_config("inventel_wireless_magic", &inventel_wireless_magic, CONFIG_INT) == false) {
        printf("1\n");
        return 1;
      }
      if ((inventel_wireless_magic & 0x00000001) == 0x00000001) {
        printf("1\n");
      }
      else {
        printf("0\n");
      }
    }


#if defined CONFIG_SUPPORT_EXOTIC_PASSWD_AFONE
    else if (strcmp(parameter, "af_adsl_pass") == 0 ) {

      /* passwd adsl afone : les huits premiers caractres du paramtre h235_key_data */

      flash_get_config("h235_key", &h235_key, CONFIG_BOOL);
      diag_printf("H235 KEY : ");
      if (h235_key) {
	if (flash_get_config("h235_key_data", &h235_key_dyn, CONFIG_STRING) == false) {
	  printf("\n");
	}
	else {
	  char afone_adsl_pass [9] ;
	  strncpy (afone_adsl_pass, h235_key_dyn, 8);
	  afone_adsl_pass[8] = 0;
	  printf("%s\n",afone_adsl_pass );
	}
      }
    }
    else if (strcmp(parameter, "af_gwname") == 0 ) {
      /* GW NAME afone : 
       * 8 premiers octets de MD5SUM de "af_adsl_pass" + "Inventel 2005" + "b4pPVcGrg8Lz"
       * 3 octets unique ( = 3 derniers octets mac XOR 0x494E56 )
       * + 9 octets de MD5SUM de "af_adsl_pass" + "adiozm45VIoQ"
       * -> 20  octets -> base32 -> 30 premiers caractres 
       */

      flash_get_config("h235_key", &h235_key, CONFIG_BOOL);

      if (h235_key) {
	diag_printf(" - ");
	if (flash_get_config("h235_key_data", &h235_key_dyn, CONFIG_STRING) == false) {
	  diag_printf (" NOK ");
	  printf("\n");
	}
	else {
	  char afone_adsl_pass [9] ;
	  char tableau[32];
	  char out_tableau[33];
	  FILE * pfile;
	  FILE * ffile;
	  int c; 
	  int i=0;
	  
	  strncpy (afone_adsl_pass, h235_key_dyn, 8);
	  afone_adsl_pass[8] = 0;
	  pfile = popen ("md5sum > /var/tmp/~md","w");
	  if (! pfile) {
	    diag_printf("erreur md5 1 \n");
	    return 1;
	  }
	  fprintf(pfile, "%s",afone_adsl_pass);
	  fprintf(pfile, "%s","Inventel 2005");
	  fprintf(pfile, "%s","b4pPVcGrg8Lz");

	  pclose (pfile);
	  ffile = fopen ("/var/tmp/~md", "r");
	  if (!ffile) {
	    diag_printf("erreur fichier\n")
	    return 1;
	  }
	  while ((c = fgetc(ffile)) > 0 ) {
	    
	    diag_printf("c : %c\n",c);
	    if (((c<'0') || (c>'9')) && ((c<'a') || (c>'f'))) {
	      diag_printf("mauvais caractere : %c\n",c);
	      continue;
	    }

	    tableau[i]=((c > '9') ?  (c - 'a' + 10 ) : (c - '0'))<<4;
	    diag_printf("%x\n",tableau[i]);

	    if (( c = fgetc (ffile))>0) 
	      tableau[i]|= ((c > '9') ?  (c - 'a' + 10 ) : (c - '0'))  ;
	    diag_printf("%x\n",tableau[i]);
	    i ++;
	  }
	  fclose (ffile);
	  if (i!=16) {
	    diag_printf("erreur taille : %d \n",i);
	    return 1;
	  }
	  i = 8; 

	  flash_get_config("eth0_esa", &set_esa, CONFIG_BOOL);	  if (set_esa) {
	    if (flash_get_config("eth0_esa_data", esa, CONFIG_ESA) == false) {
	      printf("00:07:3A:FF:FF:FF\n");
	    }
	    else {
	      diag_printf("3 caracteres addr mac : 0x%02x 0x%02x 0x%02x \n",
			  esa[3],esa[4],esa[5]);
	    }
	  }
	  
	  tableau[i++] = esa[3] ^ 'I';
	  tableau[i++] = esa[4] ^ 'N';
	  tableau[i++] = esa[5] ^ 'V';

	  pfile = popen ("md5sum > /var/tmp/~md2","w");
	  if (! pfile) {
	    diag_printf("erreur md5 2 \n");
	    return 1;
	  }
	  fprintf(pfile, "%s",afone_adsl_pass);
	  fprintf(pfile, "%s","adiozm45VIoQ");	  

	  pclose (pfile);
	  ffile = fopen ("/var/tmp/~md2", "r");
	  if (!ffile) { 
	    diag_printf("erreur fichier 2\n");
	    return 1;
	  }
	  while ((c = fgetc(ffile)) > 0 ) {

	    if (((c<'0') || (c>'9')) && ((c<'a') || (c>'f'))) continue;

	    tableau[i]=((c > '9') ?  (c - 'a' +10 ) : (c - '0'))<<4;
	    if (( c = fgetc (ffile))>0) 
	      tableau[i]|= ((c > '9') ?  (c - 'a' +10  ) : (c - '0')) ;
	    i ++;
	  }
	  fclose (ffile);
	  if (i!=27) return;
	  
	  /* Conversion table.  for base 32 */
	  static char b32_alphabet[32] = "abcdefghijklmnopqrstuvwxyz234567";

	  unsigned char * s = tableau;
	  unsigned char * p = out_tableau;
	  /* Transform the 5x8 bits to 8x5 bits, as required by base32.  */

	  for (i = 0; i < 20; i += 5) {
	    diag_printf("%02x %02x %02x %02x %02x |\n", s[0],s[1], s[2],s[3],s[4]);
	    *p++ = b32_alphabet[s[0] >> 3]; // caractre 0 
	    *p++ = b32_alphabet[((s[0] & 7) << 2) + (s[1] >> 6)]; // caractre 1 
	    *p++ = b32_alphabet[(s[1]>>1) & 0x1F ]; // caractre 2
	    *p++ = b32_alphabet[((s[1] & 0x1)<<4) + (s[2]>>4) ]; // caractre 3
	    *p++ = b32_alphabet[((s[2] & 0xF)<<1) + ((s[3]&0x80)>>7) ];// caractre 4
	    *p++ = b32_alphabet[(s[3] >> 2) & 0x1F] ;// caractre 5
	    *p++ = b32_alphabet[((s[3] & 0x3)<<3) + (s[4]>>5) ];// caractre 6
	    *p++ = b32_alphabet[s[4] & 0x1F];// caractre 7
	    s += 5;
	  }

	  out_tableau[30] = 0;
	  printf("%s", out_tableau);
	}
      }
    }
#endif // defined CONFIG_SUPPORT_EXOTIC_PASSWD_AFONE
    else if (strcmp(parameter, "bt_pin") == 0) {
      flash_get_config("bt_pin", &h235_key, CONFIG_BOOL);
      diag_printf("BT PIN : ");
      if (h235_key) {
	if (flash_get_config("bt_pin_data", &h235_key_dyn, CONFIG_STRING) == false) {
	  printf("0000\n");
	}
	else {
	  printf("%s\n", h235_key_dyn);
	}
      }
      else {
	printf("0000\n");
      }
    }
    else
      diag_printf("no %s parameter\n", parameter);
    
  }
  else {
    if (strcmp(parameter, "af_sslkey") == 0 ) {
      char buf[1300];
      FILE * fd;
      int res,len;
      char * p;

      bzero(buf, 1300);

      /* RECUPERATION DE LA CLEF SSL AFONE */
      fd = fopen(device, "r");
      if (!fd) {
	diag_printf("unable to open file : %s\n", device);
	return 1;
      }
      diag_printf("open device OK : %s\n", device);
      if (fseek(fd, -2*1024,SEEK_END)) { /* key is stored in the last 2k of this flash device  */
	diag_printf("erreur fseek, %s", strerror(errno));
	return; 
      } 
      res = fread(buf, 1, 1300, fd);
      fclose(fd);
      if (res!=1300) {
	diag_printf("short read : %d\n", res);
	return 1 ; 
      } else 
	diag_printf("read : %d\n", res);

      fd = fopen("/var/tmp/key.tmp1", "w");
      if (!fd) {
	diag_printf("unable to open file : /var/tmp/key.tmp1\n", device);
	return 1;     
      }
      fwrite (buf, 1300,1, fd);
      fclose(fd);    

      p = memmem (buf, 1300, "inventel2005paris" , strlen ("inventel2005paris"));
      if (p) {
	len = p - buf;
      } else {
	diag_printf("impossible de trouver la bonne chaine\n");
	return 1;
      }
	
      fd = fopen("/var/tmp/key.tmp", "w");
      if (!fd) {
	diag_printf("unable to open file : /var/tmp/key.tmp\n", device);
	return 1;     
      }
      fwrite (buf, len,1, fd);
      fclose(fd); 
      unlink( "/var/tmp/key.tmp1");
    } else { 
      printf("No configuration avalaible\n");
      return 1;
    }
  }
  return 0;
}
