/* dhcpd.c
 *
 * Moreton Bay DHCP Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>
#include <getopt.h>

#include "debug.h"
#include "dhcpd.h"
#include "arpping.h"
#include "socket.h"
#include "options.h"
#include "files.h"
#include "leases.h"
#include "packet.h"
#include "serverpacket.h"

/* prototypes */
static int log_pid(void);


/* globals */
struct dhcpOfferedAddr *leases;
struct server_config_t server_config;

struct config_server_t config_server = {
	/* Default options. */
	config_file: DHCPD_CONF_FILE,
};

static void print_usage(void)
{
	printf(
"Usage: udhcpd [OPTIONS]\n\n"
"  -f, --config_file=file            udhcpd conf file\n"
	);
}

static void udhcpd_killed(int pid)
{
	pid = 0;
	if (server_config.pid_file) unlink(server_config.pid_file);
	LOG(LOG_INFO, "Received SIGTERM");
	CLOSE_LOG();
	exit(0);
}

int attach_interface(char *interface)
{
  int fd = -1;
  struct ifreq ifr;
  struct sockaddr_in *sin;

  if ((server_config.ifnb+1) >= MAX_INTERFACE) {
    LOG(LOG_INFO, "Too many network interfaces already(%d)", (server_config.ifnb+1));
    return -1;
  }

  server_config.interface[server_config.ifnb] = interface;

  if((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) {
    ifr.ifr_addr.sa_family = AF_INET;
    strcpy(ifr.ifr_name, interface);
    if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) {
      sin = (struct sockaddr_in *) &ifr.ifr_addr;
      server_config.server[server_config.ifnb] = sin->sin_addr.s_addr;
      DEBUG(LOG_INFO, "%s (server_ip %d) = %s", ifr.ifr_name, server_config.ifnb, inet_ntoa(sin->sin_addr));
    } else {
      LOG(LOG_ERR, "SIOCGIFADDR failed!");
      return 1;
    }
    if (ioctl(fd, SIOCGIFINDEX, &ifr) == 0) {
      DEBUG(LOG_INFO, "adapter index %d", ifr.ifr_ifindex);
      server_config.ifindex[server_config.ifnb] = ifr.ifr_ifindex;
    } else {
      LOG(LOG_ERR, "SIOCGIFINDEX failed!");
      return 1;
    }
    if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0) {
      memcpy(server_config.arp[server_config.ifnb], ifr.ifr_hwaddr.sa_data, 6);
      DEBUG(LOG_INFO, "adapter hardware address %02x:%02x:%02x:%02x:%02x:%02x",
	    server_config.arp[server_config.ifnb][0], server_config.arp[server_config.ifnb][1], server_config.arp[server_config.ifnb][2], 
	    server_config.arp[server_config.ifnb][3], server_config.arp[server_config.ifnb][4], server_config.arp[server_config.ifnb][5]);
		} else {
		  LOG(LOG_ERR, "SIOCGIFHWADDR failed!");
		  return 1;
		}
  } else {
    LOG(LOG_ERR, "socket failed!");
    return 1;
  }
  close(fd);

  server_config.ifnb++;
  
  return 0;
}

#ifdef COMBINED_BINARY	
int udhcpd(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{	
        /* CMO */
        unsigned char * hostname;
	unsigned int hostname_len;
	fd_set rfds;
	struct timeval tv;
	int server_socket[MAX_INTERFACE];
	int bytes, retval;
	struct dhcpMessage packet;
	unsigned char *state;
	char *server_id, *requested;
	u_int32_t server_id_align, requested_align;
	unsigned long timeout_end;
	struct option_set *option;
	struct dhcpOfferedAddr *lease;
	//	struct sockaddr_in *sin;
			
	/* server ip addr */
	//int fd = -1;
	//struct ifreq ifr;

	static struct option options[] = {
		{"config_file",	required_argument,		0, 'f'},
		{"interface",	required_argument,		0, 'i'},
		{"help",	no_argument,		0, 'h'},
		{0, 0, 0, 0}
	};
  
        int c ;

        memset(&server_config, 0, sizeof(struct server_config_t));

	/* get options */
	while (1) {
		int option_index = 0;
		c = getopt_long(argc, argv, "f:i:h", options, &option_index);
		if (c == -1) break;
		
		switch (c) {
		case 'f':
			config_server.config_file = optarg;
			printf("config_file : %s\n", config_server.config_file) ;
			break;
		case 'i':
		        attach_interface(optarg);
		        break;
		case 'h':
			print_usage();
			return 0;
		}
	}

//	argc = argv[0][0]; /* get rid of some warnings */
	
	OPEN_LOG("udhcpd");
	LOG(LOG_INFO, "Moreton Bay DHCP Server (v%s) started", VERSION);
	
//	read_config(DHCPD_CONF_FILE);
	read_config(config_server.config_file);

	if (server_config.ifnb == 0) {
	  LOG(LOG_ERR, "No interface found in configuration file");
	  CLOSE_LOG();
	  exit(0);
	}

	if ((option = find_option(server_config.options, DHCP_LEASE_TIME))) {
		memcpy(&server_config.lease, option->data + 2, 4);
		server_config.lease = ntohl(server_config.lease);
	}
	else server_config.lease = LEASE_TIME;
	
	leases = malloc(sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	memset(leases, 0, sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	read_leases(server_config.lease_file);

	log_pid();
	
	/* if((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) { */
/* 		ifr.ifr_addr.sa_family = AF_INET; */
/* 		strcpy(ifr.ifr_name, server_config.interface); */
/* 		if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) { */
/* 			sin = (struct sockaddr_in *) &ifr.ifr_addr; */
/* 			server_config.server = sin->sin_addr.s_addr; */
/* 			DEBUG(LOG_INFO, "%s (server_ip) = %s", ifr.ifr_name, inet_ntoa(sin->sin_addr)); */
/* 		} else { */
/* 			LOG(LOG_ERR, "SIOCGIFADDR failed!"); */
/* 			return 1; */
/* 		} */
/* 		if (ioctl(fd, SIOCGIFINDEX, &ifr) == 0) { */
/* 			DEBUG(LOG_INFO, "adapter index %d", ifr.ifr_ifindex); */
/* 			server_config.ifindex = ifr.ifr_ifindex; */
/* 		} else { */
/* 			LOG(LOG_ERR, "SIOCGIFINDEX failed!"); */
/* 			return 1; */
/* 		} */
/* 		if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0) { */
/* 			memcpy(server_config.arp, ifr.ifr_hwaddr.sa_data, 6); */
/* 			DEBUG(LOG_INFO, "adapter hardware address %02x:%02x:%02x:%02x:%02x:%02x", */
/* 				server_config.arp[0], server_config.arp[1], server_config.arp[2],  */
/* 				server_config.arp[3], server_config.arp[4], server_config.arp[5]); */
/* 		} else { */
/* 			LOG(LOG_ERR, "SIOCGIFHWADDR failed!"); */
/* 			return 1; */
/* 		} */
/* 	} else { */
/* 		LOG(LOG_ERR, "socket failed!"); */
/* 		return 1; */
/* 	} */
/* 	close(fd); */

#ifndef DEBUGGING
	if (fork()) return 0;
	else {
		close(0);
		setsid();
	}
#endif


	signal(SIGUSR1, write_leases);
	signal(SIGTERM, udhcpd_killed);

	timeout_end = time(0) + server_config.auto_time;
	while(1) { /* loop until universe collapses */
	  int sock = 0, i;
	  u_int8_t *clientid = 0;
	  
		FD_ZERO(&rfds);

		for (i=0; i<server_config.ifnb; i++) {

		  server_socket[i] = listen_socket(INADDR_ANY, SERVER_PORT, server_config.interface[i]);
		  if(server_socket[i] == -1) {
		    LOG(LOG_ERR, "couldn't create server socket -- au revoir");
		    exit(0);
		  }			
		  
		  if (server_socket[i] > sock)
		    sock = server_socket[i];

		  FD_SET(server_socket[i], &rfds);
		}

		if (server_config.auto_time) {
			tv.tv_sec = timeout_end - time(0);
			if (tv.tv_sec <= 0) {
				tv.tv_sec = server_config.auto_time;
				timeout_end = time(0) + server_config.auto_time;
				write_leases(0);
			}
			tv.tv_usec = 0;
		}

		retval = select(sock + 1, &rfds, NULL, NULL, server_config.auto_time ? &tv : NULL);
		if (retval == 0) {
			write_leases(0);
			timeout_end = time(0) + server_config.auto_time;
			for (i=0; i<server_config.ifnb; i++)
			  close(server_socket[i]);
			continue;
		} else if (retval < 0) {
			DEBUG(LOG_INFO, "error on select");
			for (i=0; i<server_config.ifnb; i++)
			  close(server_socket[i]);
			continue;
		}

		sock = 0;
		for (i=0; i<server_config.ifnb; i++) {
		  if ( FD_ISSET(server_socket[i], &rfds) ) {
		    DEBUG(LOG_INFO, "got something on server_socket[%d]",i);
		    sock = server_socket[i] ;
		    server_config.selectedif = i;
		    break;
		  }
		}
		if (sock == 0) {
		  for (i=0; i<server_config.ifnb; i++)
		    close(server_socket[i]);
		  continue;
		}

		bytes = get_packet(&packet, sock); /* this waits for a packet - idle */
		for (i=0; i<server_config.ifnb; i++)
		  close(server_socket[i]);
		if(bytes < 0)
			continue;

		if((state = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
			DEBUG(LOG_ERR, "couldnt get option from packet -- ignoring");
			continue;
		}

		/* CMO */
		if((hostname = get_option(&packet, DHCP_HOST_NAME))) {
		  
		  hostname_len= (unsigned char) *(hostname - 1); /* hack : la longueur est l'octet avant les dones */
		  DEBUG(LOG_ERR, "HOSTNAME extraction : %.*s\n",hostname_len,hostname);
		} else {
		  hostname_len=0;
		  DEBUG(LOG_ERR, "NO HOSTNAME \n");
		}

		/* fin CMO */

		if ( ((clientid = get_option(&packet, DHCP_CLIENT_ID)) != NULL) && (*clientid == 1) ) {
		  // a client identifier as an eth hw address is given
		  memset(server_config.client_chaddr,0,16);
		  memcpy(server_config.client_chaddr,clientid+1,6);
		  DEBUG(LOG_INFO,"received DHCP_CLIENT_ID %02x:%02x:%02x:%02x:%02x:%02x", server_config.client_chaddr[0], server_config.client_chaddr[1], server_config.client_chaddr[2], server_config.client_chaddr[3], server_config.client_chaddr[4], server_config.client_chaddr[5]);
		} else {
		  memcpy(server_config.client_chaddr,packet.chaddr,16);
		}

		lease = find_lease_by_chaddr(server_config.client_chaddr);
		switch (state[0]) {
		case DHCPDISCOVER:
			DEBUG(LOG_INFO,"received DISCOVER");
			sleep(1);
			if (sendOffer(&packet) < 0) {
				LOG(LOG_ERR, "send OFFER failed -- ignoring");
			}
			break;			
 		case DHCPREQUEST:
			DEBUG(LOG_INFO,"received REQUEST");
			sleep(1);
			requested = get_option(&packet, DHCP_REQUESTED_IP);
			server_id = get_option(&packet, DHCP_SERVER_ID);

			if (requested) memcpy(&requested_align, requested, 4);
			if (server_id) memcpy(&server_id_align, server_id, 4);
		
			if (lease) {
			        int was_ok=0;
				if (server_id) {
					/* SELECTING State */
				 
					if (server_id_align == server_config.server[server_config.selectedif] && requested && 
					    requested_align == lease->yiaddr) {					  
						sendACK(&packet, lease->yiaddr);
						was_ok=1;
					}
				} else {
				        
					if (requested) {
						/* INIT-REBOOT State */
				
					  if (lease->yiaddr == requested_align) {
					    sendACK(&packet, lease->yiaddr); was_ok=1;
					  }
						else sendNAK(&packet);
					} else {
						/* RENEWING or REBINDING State */
					  if (lease->yiaddr == packet.ciaddr) {
					    sendACK(&packet, lease->yiaddr);
					    was_ok=1;
					  }
						else {
						  /* don't know what to do!!!! */
							sendNAK(&packet);
						}
					}						
				}
				/* CMO : maintenant, nous avons le temps de changer la liste des priphriques associs 
				 * pour mettre le hostname dans le champs nom. 
				 */
				if (was_ok && hostname_len) {
				  FILE * fr=NULL ;
				  FILE * fw=NULL ;
//				  publish_hostname(server_config.client_chaddr, hostname, hostname_len); 
//				  sprintf(cmd, "sed \"s;%02x:%02x:%02x:%02x:%02x:%02x.*;%02x:%02x:%02x:%02x:%02x:%02x %.*s \" /etc/wifi_mac > /etc/wifi_mac.tmp
				  if (!(fr = fopen ("/etc/wifi_mac", "r"))) 
				    goto out;
				  if (!(fw = fopen ("/etc/wifi_mac.tmp", "w")))
				    goto out;
				  
				  
				  DEBUG(LOG_INFO, "HOSTNAME extraction : %.*s, addresse mac : %02x:%02x:%02x:%02x:%02x:%02x\n",hostname_len,hostname, 
					server_config.client_chaddr[0], 
					server_config.client_chaddr[1], 
					server_config.client_chaddr[2], 
					server_config.client_chaddr[3], 
					server_config.client_chaddr[4], 
					server_config.client_chaddr[5]);
				out:
				  if (fr) fclose (fr);
				  if (fw) fclose (fw);
				}
			} /* else remain silent */
			break;
		case DHCPDECLINE:
			DEBUG(LOG_INFO,"received DECLINE");
			if (lease) {
				memset(lease->chaddr, 0, 16);
				lease->expires = time(0) + server_config.decline_time;
			}			
			break;
		case DHCPRELEASE:
			DEBUG(LOG_INFO,"received RELEASE");
			if (lease) lease->expires = time(0);
			break;
		case DHCPINFORM:
			DEBUG(LOG_INFO,"received INFORM");
			send_inform(&packet);
			break;	
		default:
			LOG(LOG_WARNING, "unsupported DHCP message (%02x) -- ignoring", state[0]);
		}
	}

	return 0;
}


int log_pid(void) 
{
/* xde */
/*	int fd;*/
	pid_t pid;
	char *pidfile = server_config.pid_file;
  FILE *file_ptr ;

	pid = getpid();
/* xde */
/*	if((fd = open(pidfile, O_WRONLY | O_CREAT, 0660)) < 0)*/
	if((file_ptr = fopen(pidfile, "w")) == NULL)
  {
    printf("Unable to open %s\n", pidfile) ;
		return -1;
  }
/* xde */
/*	write(fd, (void *) &pid, sizeof(pid));*/
	fprintf(file_ptr, "%d\n", (int)pid);
/* xde */
/*	close(fd);*/
	fclose(file_ptr);
	return 0;
}


